"""Core settings."""

from collections import OrderedDict

import collections

from django import forms
from django.conf import settings
from django.contrib.auth import password_validation
from django.utils.translation import ugettext as _, ugettext_lazy

from modoboa.core.password_hashers import get_dovecot_schemes
from modoboa.core.password_hashers.base import PasswordHasher
from modoboa.lib import fields as lib_fields
from modoboa.lib.form_utils import (
    HorizontalRadioSelect, SeparatorField, YesNoField
)
from modoboa.parameters import forms as param_forms, tools as param_tools

from . import constants
from . import sms_backends


def enabled_applications():
    """Return the list of installed extensions."""
    from modoboa.core.extensions import exts_pool

    result = [("user", _("User profile"))]
    for extension in exts_pool.list_all():
        if "topredirection_url" not in extension:
            continue
        result.append((extension["name"], extension["label"]))
    return sorted(result, key=lambda e: e[0])


class GeneralParametersForm(param_forms.AdminParametersForm):
    """General parameters."""

    app = "core"

    sep1 = SeparatorField(label=ugettext_lazy("Authentication"))

    authentication_type = forms.ChoiceField(
        label=ugettext_lazy("Authentication type"),
        choices=[("local", ugettext_lazy("Local")),
                 ("ldap", "LDAP")],
        initial="local",
        help_text=ugettext_lazy("The backend used for authentication"),
        widget=HorizontalRadioSelect()
    )

    password_scheme = forms.ChoiceField(
        label=ugettext_lazy("Default password scheme"),
        choices=[],
        initial="sha512crypt",
        help_text=ugettext_lazy("Scheme used to crypt mailbox passwords"),
    )

    rounds_number = forms.IntegerField(
        label=ugettext_lazy("Rounds"),
        initial=70000,
        help_text=ugettext_lazy(
            "Number of rounds to use (only used by sha256crypt and "
            "sha512crypt). Must be between 1000 and 999999999, inclusive."
        ),
    )

    update_scheme = YesNoField(
        label=ugettext_lazy("Update password scheme at login"),
        initial=True,
        help_text=ugettext_lazy(
            "Update user password at login to use the default password scheme"
        )
    )

    default_password = forms.CharField(
        label=ugettext_lazy("Default password"),
        initial="password",
        help_text=ugettext_lazy(
            "Default password for automatically created accounts.")
    )

    random_password_length = forms.IntegerField(
        label=ugettext_lazy("Random password length"),
        min_value=8,
        initial=8,
        help_text=ugettext_lazy(
            "Length of randomly generated passwords.")
    )

    update_password_url = forms.URLField(
        label=ugettext_lazy("Update password service URL"),
        initial="",
        required=False,
        help_text=ugettext_lazy(
            "The URL of an external page where users will be able"
            " to update their password. It applies only to non local"
            " users, ie. those automatically created after a successful"
            " external authentication (LDAP, SMTP)."
        )
    )

    password_recovery_msg = forms.CharField(
        label=ugettext_lazy("Password recovery announcement"),
        initial="",
        required=False,
        widget=forms.widgets.Textarea(),
        help_text=ugettext_lazy(
            "A temporary message that will be displayed on the "
            "reset password page."
        )
    )

    sms_password_recovery = YesNoField(
        label=ugettext_lazy("Enable password recovery by SMS"),
        initial=False,
        help_text=ugettext_lazy(
            "Enable password recovery by SMS for users who filled "
            "a phone number."
        )
    )

    sms_provider = forms.ChoiceField(
        label=ugettext_lazy("SMS provider"),
        choices=constants.SMS_BACKENDS,
        help_text=ugettext_lazy(
            "Choose a provider to send password recovery SMS"
        ),
        required=False
    )

    # LDAP specific settings
    ldap_sep = SeparatorField(label=ugettext_lazy("LDAP settings"))

    ldap_server_address = forms.CharField(
        label=ugettext_lazy("Server address"),
        initial="localhost",
        help_text=ugettext_lazy(
            "The IP address or the DNS name of the LDAP server"),
    )

    ldap_server_port = forms.IntegerField(
        label=ugettext_lazy("Server port"),
        initial=389,
        help_text=ugettext_lazy("The TCP port number used by the LDAP server"),
    )

    ldap_enable_secondary_server = YesNoField(
        label=ugettext_lazy("Enable secondary server (fallback)"),
        initial=False,
        help_text=ugettext_lazy(
            "Enable a secondary LDAP server which will be used "
            "if the primary one fails"
        )
    )

    ldap_secondary_server_address = forms.CharField(
        label=ugettext_lazy("Secondary server address"),
        initial="localhost",
        help_text=ugettext_lazy(
            "The IP address or the DNS name of the seondary LDAP server"),
    )

    ldap_secondary_server_port = forms.IntegerField(
        label=ugettext_lazy("Secondary server port"),
        initial=389,
        help_text=ugettext_lazy(
            "The TCP port number used by the LDAP secondary server"),
    )

    ldap_secured = forms.ChoiceField(
        label=ugettext_lazy("Use a secured connection"),
        choices=constants.LDAP_SECURE_MODES,
        initial="none",
        help_text=ugettext_lazy(
            "Use an SSL/STARTTLS connection to access the LDAP server")
    )

    ldap_is_active_directory = YesNoField(
        label=ugettext_lazy("Active Directory"),
        initial=False,
        help_text=ugettext_lazy(
            "Tell if the LDAP server is an Active Directory one")
    )

    ldap_admin_groups = forms.CharField(
        label=ugettext_lazy("Administrator groups"),
        initial="",
        help_text=ugettext_lazy(
            "Members of those LDAP Posix groups will be created as domain "
            "administrators. Use ';' characters to separate groups."
        ),
        required=False
    )

    ldap_group_type = forms.ChoiceField(
        label=ugettext_lazy("Group type"),
        initial="posixgroup",
        choices=constants.LDAP_GROUP_TYPES,
        help_text=ugettext_lazy(
            "The LDAP group type to use with your directory."
        )
    )

    ldap_groups_search_base = forms.CharField(
        label=ugettext_lazy("Groups search base"),
        initial="",
        help_text=ugettext_lazy(
            "The distinguished name of the search base used to find groups"
        ),
        required=False
    )

    ldap_password_attribute = forms.CharField(
        label=ugettext_lazy("Password attribute"),
        initial="userPassword",
        help_text=ugettext_lazy("The attribute used to store user passwords"),
    )

    # LDAP authentication settings
    ldap_auth_sep = SeparatorField(
        label=ugettext_lazy("LDAP authentication settings"))

    ldap_auth_method = forms.ChoiceField(
        label=ugettext_lazy("Authentication method"),
        choices=[("searchbind", ugettext_lazy("Search and bind")),
                 ("directbind", ugettext_lazy("Direct bind"))],
        initial="searchbind",
        help_text=ugettext_lazy("Choose the authentication method to use"),
    )

    ldap_bind_dn = forms.CharField(
        label=ugettext_lazy("Bind DN"),
        initial="",
        help_text=ugettext_lazy(
            "The distinguished name to use when binding to the LDAP server. "
            "Leave empty for an anonymous bind"
        ),
        required=False,
    )

    ldap_bind_password = forms.CharField(
        label=ugettext_lazy("Bind password"),
        initial="",
        help_text=ugettext_lazy(
            "The password to use when binding to the LDAP server "
            "(with 'Bind DN')"
        ),
        widget=forms.PasswordInput(render_value=True),
        required=False
    )

    ldap_search_base = forms.CharField(
        label=ugettext_lazy("Users search base"),
        initial="",
        help_text=ugettext_lazy(
            "The distinguished name of the search base used to find users"
        ),
        required=False,
    )

    ldap_search_filter = forms.CharField(
        label=ugettext_lazy("Search filter"),
        initial="(mail=%(user)s)",
        help_text=ugettext_lazy(
            "An optional filter string (e.g. '(objectClass=person)'). "
            "In order to be valid, it must be enclosed in parentheses."
        ),
        required=False,
    )

    ldap_user_dn_template = forms.CharField(
        label=ugettext_lazy("User DN template"),
        initial="",
        help_text=ugettext_lazy(
            "The template used to construct a user's DN. It should contain "
            "one placeholder (ie. %(user)s)"
        ),
        required=False,
    )

    # LDAP sync. settings
    ldap_sync_sep = SeparatorField(
        label=ugettext_lazy("LDAP synchronization settings"))

    ldap_sync_bind_dn = forms.CharField(
        label=ugettext_lazy("Bind DN"),
        initial="",
        help_text=ugettext_lazy(
            "The distinguished name to use when binding to the LDAP server. "
            "Leave empty for an anonymous bind"
        ),
        required=False,
    )

    ldap_sync_bind_password = forms.CharField(
        label=ugettext_lazy("Bind password"),
        initial="",
        help_text=ugettext_lazy(
            "The password to use when binding to the LDAP server "
            "(with 'Bind DN')"
        ),
        widget=forms.PasswordInput(render_value=True),
        required=False
    )

    ldap_enable_sync = YesNoField(
        label=ugettext_lazy("Enable export to LDAP"),
        initial=False,
        help_text=ugettext_lazy(
            "Enable automatic synchronization between local database and "
            "LDAP directory")
    )

    ldap_sync_delete_remote_account = YesNoField(
        label=ugettext_lazy(
            "Delete remote LDAP account when local account is deleted"
        ),
        initial=False,
        help_text=ugettext_lazy(
            "Delete remote LDAP account when local account is deleted, "
            "otherwise it will be disabled."
        )
    )

    ldap_sync_account_dn_template = forms.CharField(
        label=ugettext_lazy("Account DN template"),
        initial="",
        help_text=ugettext_lazy(
            "The template used to construct an account's DN. It should contain "
            "one placeholder (ie. %(user)s)"
        ),
        required=False
    )

    ldap_enable_import = YesNoField(
        label=ugettext_lazy("Enable import from LDAP"),
        initial=False,
        help_text=ugettext_lazy(
            "Enable account synchronization from LDAP directory to local "
            "database"
        )
    )

    ldap_import_search_base = forms.CharField(
        label=ugettext_lazy("Users search base"),
        initial="",
        help_text=ugettext_lazy(
            "The distinguished name of the search base used to find users"
        ),
        required=False,
    )

    ldap_import_search_filter = forms.CharField(
        label=ugettext_lazy("Search filter"),
        initial="(cn=*)",
        help_text=ugettext_lazy(
            "An optional filter string (e.g. '(objectClass=person)'). "
            "In order to be valid, it must be enclosed in parentheses."
        ),
        required=False,
    )

    ldap_import_username_attr = forms.CharField(
        label=ugettext_lazy("Username attribute"),
        initial="cn",
        help_text=ugettext_lazy(
            "The name of the LDAP attribute where the username can be found."
        ),
    )

    ldap_dovecot_sync = YesNoField(
        label=ugettext_lazy("Enable Dovecot LDAP sync"),
        initial=False,
        help_text=ugettext_lazy(
            "LDAP authentication settings will be applied to Dovecot "
            "configuration."
        )
    )

    ldap_dovecot_conf_file = forms.CharField(
        label=ugettext_lazy("Dovecot LDAP config file"),
        initial="/etc/dovecot/dovecot-modoboa.conf",
        required=False,
        help_text=ugettext_lazy(
            "Location of the configuration file which contains "
            "Dovecot LDAP settings."
        )
    )

    dash_sep = SeparatorField(label=ugettext_lazy("Dashboard"))

    rss_feed_url = forms.URLField(
        label=ugettext_lazy("Custom RSS feed"),
        required=False,
        help_text=ugettext_lazy(
            "Display custom RSS feed to resellers and domain administrators"
        )
    )

    hide_features_widget = YesNoField(
        label=ugettext_lazy("Hide features widget"),
        initial=False,
        help_text=ugettext_lazy(
            "Hide features widget for resellers and domain administrators"
        )
    )

    notif_sep = SeparatorField(label=ugettext_lazy("Notifications"))

    sender_address = lib_fields.UTF8EmailField(
        label=_("Sender address"),
        initial="noreply@yourdomain.test",
        help_text=_(
            "Email address used to send notifications."
        )
    )

    api_sep = SeparatorField(label=ugettext_lazy("Public API"))

    enable_api_communication = YesNoField(
        label=ugettext_lazy("Enable communication"),
        initial=True,
        help_text=ugettext_lazy(
            "Enable communication with Modoboa public API")
    )

    check_new_versions = YesNoField(
        label=ugettext_lazy("Check new versions"),
        initial=True,
        help_text=ugettext_lazy(
            "Automatically checks if a newer version is available")
    )

    send_new_versions_email = YesNoField(
        label=ugettext_lazy("Send an email when new versions are found"),
        initial=False,
        help_text=ugettext_lazy(
            "Send an email to notify admins about new versions"
        )
    )
    new_versions_email_rcpt = lib_fields.UTF8EmailField(
        label=_("Recipient"),
        initial="postmaster@yourdomain.test",
        help_text=_(
            "Recipient of new versions notification emails."
        )
    )

    send_statistics = YesNoField(
        label=ugettext_lazy("Send statistics"),
        initial=True,
        help_text=ugettext_lazy(
            "Send statistics to Modoboa public API "
            "(counters and used extensions)")
    )

    sep3 = SeparatorField(label=ugettext_lazy("Miscellaneous"))

    inactive_account_threshold = forms.IntegerField(
        label=_("Inactive account threshold"),
        initial=30,
        help_text=_(
            "An account with a last login date greater than this threshold "
            "(in days) will be considered as inactive"
        ),
    )

    top_notifications_check_interval = forms.IntegerField(
        label=_("Top notifications check interval"),
        initial=30,
        help_text=_(
            "Interval between two top notification checks (in seconds)"
        ),
    )

    log_maximum_age = forms.IntegerField(
        label=ugettext_lazy("Maximum log record age"),
        initial=365,
        help_text=ugettext_lazy("The maximum age in days of a log record"),
    )

    items_per_page = forms.IntegerField(
        label=ugettext_lazy("Items per page"),
        initial=30,
        help_text=ugettext_lazy("Number of displayed items per page"),
    )

    default_top_redirection = forms.ChoiceField(
        label=ugettext_lazy("Default top redirection"),
        choices=[],
        initial="user",
        help_text=ugettext_lazy(
            "The default redirection used when no application is specified"
        ),
    )

    # Visibility rules
    visibility_rules = {
        "ldap_secondary_server_address": "ldap_enable_secondary_server=True",
        "ldap_secondary_server_port": "ldap_enable_secondary_server=True",
        "ldap_auth_sep": "authentication_type=ldap",
        "ldap_auth_method": "authentication_type=ldap",
        "ldap_bind_dn": "ldap_auth_method=searchbind",
        "ldap_bind_password": "ldap_auth_method=searchbind",
        "ldap_search_base": "ldap_auth_method=searchbind",
        "ldap_search_filter": "ldap_auth_method=searchbind",
        "ldap_user_dn_template": "ldap_auth_method=directbind",
        "ldap_admin_groups": "authentication_type=ldap",
        "ldap_group_type": "authentication_type=ldap",
        "ldap_groups_search_base": "authentication_type=ldap",
        "ldap_sync_delete_remote_account": "ldap_enable_sync=True",
        "ldap_sync_account_dn_template": "ldap_enable_sync=True",
        "ldap_import_search_base": "ldap_enable_import=True",
        "ldap_import_search_filter": "ldap_enable_import=True",
        "ldap_import_username_attr": "ldap_enable_import=True",
        "ldap_dovecot_conf_file": "ldap_dovecot_sync=True",
        "check_new_versions": "enable_api_communication=True",
        "send_statistics": "enable_api_communication=True",
        "send_new_versions_email": "check_new_versions=True",
        "new_versions_email_rcpt": "check_new_versions=True",
        "sms_provider": "sms_password_recovery=True",
    }

    def __init__(self, *args, **kwargs):
        super(GeneralParametersForm, self).__init__(*args, **kwargs)
        self.fields["default_top_redirection"].choices = enabled_applications()
        self._add_visibilty_rules(
            sms_backends.get_all_backend_visibility_rules()
        )
        self.fields["password_scheme"].choices = [
            (hasher.name, ugettext_lazy(hasher.label))
            for hasher in PasswordHasher.get_password_hashers()
            if hasher().scheme in get_dovecot_schemes()
        ]

    def _add_dynamic_fields(self):
        new_fields = OrderedDict()
        for field, value in self.fields.items():
            new_fields[field] = value
            if field == "sms_provider":
                sms_backend_fields = sms_backends.get_all_backend_settings()
                for field, definition in sms_backend_fields.items():
                    new_fields[field] = definition["type"](
                        **definition["attrs"])
        self.fields = new_fields

    def clean_ldap_user_dn_template(self):
        tpl = self.cleaned_data["ldap_user_dn_template"]
        try:
            tpl % {"user": "toto"}
        except (KeyError, ValueError):
            raise forms.ValidationError(_("Invalid syntax"))
        return tpl

    def clean_ldap_sync_account_dn_template(self):
        tpl = self.cleaned_data["ldap_sync_account_dn_template"]
        try:
            tpl % {"user": "toto"}
        except (KeyError, ValueError):
            raise forms.ValidationError(_("Invalid syntax"))
        return tpl

    def clean_ldap_search_filter(self):
        ldap_filter = self.cleaned_data["ldap_search_filter"]
        try:
            ldap_filter % {"user": "toto"}
        except (KeyError, ValueError, TypeError):
            raise forms.ValidationError(_("Invalid syntax"))
        return ldap_filter

    def clean_rounds_number(self):
        value = self.cleaned_data["rounds_number"]
        if value < 1000 or value > 999999999:
            raise forms.ValidationError(_("Invalid rounds number"))
        return value

    def clean_default_password(self):
        """Check password complexity."""
        value = self.cleaned_data["default_password"]
        password_validation.validate_password(value)
        return value

    def clean(self):
        """Custom validation method

        Depending on 'ldap_auth_method' value, we check for different
        required parameters.
        """
        super().clean()
        cleaned_data = self.cleaned_data

        if cleaned_data["sms_password_recovery"]:
            provider = cleaned_data.get("sms_provider")
            if provider:
                sms_settings = sms_backends.get_backend_settings(provider)
                if sms_settings:
                    for name in sms_settings.keys():
                        if not cleaned_data.get(name):
                            self.add_error(name, _("This field is required"))
            else:
                self.add_error("sms_provider", _("This field is required"))
        if cleaned_data["authentication_type"] != "ldap":
            return cleaned_data

        if cleaned_data["ldap_auth_method"] == "searchbind":
            required_fields = ["ldap_search_base", "ldap_search_filter"]
        else:
            required_fields = ["ldap_user_dn_template"]

        for f in required_fields:
            if f not in cleaned_data or cleaned_data[f] == u'':
                self.add_error(f, _("This field is required"))

        return cleaned_data

    def _apply_ldap_settings(self, values, backend):
        """Apply configuration for given backend."""
        import ldap
        from django_auth_ldap.config import (
            LDAPSearch, PosixGroupType, GroupOfNamesType,
            ActiveDirectoryGroupType
        )

        if not hasattr(settings, backend.setting_fullname("USER_ATTR_MAP")):
            setattr(settings, backend.setting_fullname("USER_ATTR_MAP"), {
                "first_name": "givenName",
                "email": "mail",
                "last_name": "sn"
            })
        ldap_uri = "ldaps://" if values["ldap_secured"] == "ssl" else "ldap://"
        ldap_uri += "%s:%s" % (
            values[backend.srv_address_setting_name],
            values[backend.srv_port_setting_name]
        )
        setattr(settings, backend.setting_fullname("SERVER_URI"), ldap_uri)
        if values["ldap_secured"] == "starttls":
            setattr(settings, backend.setting_fullname("START_TLS"), True)

        if values["ldap_is_active_directory"]:
            setattr(
                settings, backend.setting_fullname("GROUP_TYPE"),
                ActiveDirectoryGroupType()
            )
            searchfilter = "(objectClass=group)"
        elif values["ldap_group_type"] == "groupofnames":
            setattr(settings, backend.setting_fullname("GROUP_TYPE"),
                    GroupOfNamesType())
            searchfilter = "(objectClass=groupOfNames)"
        else:
            setattr(settings, backend.setting_fullname("GROUP_TYPE"),
                    PosixGroupType())
            searchfilter = "(objectClass=posixGroup)"
        setattr(settings, backend.setting_fullname("GROUP_SEARCH"), LDAPSearch(
            values["ldap_groups_search_base"], ldap.SCOPE_SUBTREE,
            searchfilter
        ))
        if values["ldap_auth_method"] == "searchbind":
            setattr(settings, backend.setting_fullname("BIND_DN"),
                    values["ldap_bind_dn"])
            setattr(
                settings, backend.setting_fullname("BIND_PASSWORD"),
                values["ldap_bind_password"]
            )
            search = LDAPSearch(
                values["ldap_search_base"], ldap.SCOPE_SUBTREE,
                values["ldap_search_filter"]
            )
            setattr(settings, backend.setting_fullname("USER_SEARCH"), search)
        else:
            setattr(
                settings, backend.setting_fullname("USER_DN_TEMPLATE"),
                values["ldap_user_dn_template"]
            )
            setattr(
                settings,
                backend.setting_fullname("BIND_AS_AUTHENTICATING_USER"), True)
        if values["ldap_is_active_directory"]:
            setting = backend.setting_fullname("GLOBAL_OPTIONS")
            if not hasattr(settings, setting):
                setattr(settings, setting, {
                    ldap.OPT_REFERRALS: False
                })
            else:
                getattr(settings, setting)[ldap.OPT_REFERRALS] = False

    def to_django_settings(self):
        """Apply LDAP related parameters to Django settings.

        Doing so, we can use the django_auth_ldap module.
        """
        try:
            import ldap
            ldap_available = True
        except ImportError:
            ldap_available = False

        values = dict(param_tools.get_global_parameters("core"))
        if not ldap_available or values["authentication_type"] != "ldap":
            return

        from modoboa.lib.authbackends import LDAPBackend
        self._apply_ldap_settings(values, LDAPBackend)

        if not values["ldap_enable_secondary_server"]:
            return

        from modoboa.lib.authbackends import LDAPSecondaryBackend
        self._apply_ldap_settings(values, LDAPSecondaryBackend)

    def save(self):
        """Extra save actions."""
        super().save()
        self.localconfig.need_dovecot_update = True
        self.localconfig.save(update_fields=["need_dovecot_update"])


GLOBAL_PARAMETERS_STRUCT = collections.OrderedDict([
    ("authentication", {
        "label": ugettext_lazy("Authentication"),
        "params": collections.OrderedDict([
            ("authentication_type", {
                "label": ugettext_lazy("Authentication type"),
                "help_text": ugettext_lazy(
                    "The backend used for authentication"),
            }),
            ("password_scheme", {
                "label": ugettext_lazy("Default password scheme"),
                "help_text": ugettext_lazy(
                    "Scheme used to crypt mailbox passwords"),
            }),
            ("rounds_number", {
                "label": ugettext_lazy("Rounds"),
                "help_text": ugettext_lazy(
                    "Number of rounds to use (only used by sha256crypt and "
                    "sha512crypt). Must be between 1000 and 999999999, "
                    "inclusive."
                )
            }),
            ("update_scheme", {
                "label": ugettext_lazy("Update password scheme at login"),
                "help_text": ugettext_lazy(
                    "Update user password at login to use the default "
                    "password scheme"
                )
            }),
            ("default_password", {
                "label": ugettext_lazy("Default password"),
                "help_text": ugettext_lazy(
                    "Default password for automatically created accounts.")
            }),
            ("random_password_length", {
                "label": ugettext_lazy("Random password length"),
                "help_text": ugettext_lazy(
                    "Length of randomly generated passwords.")
            }),
            ("update_password_url", {
                "label": ugettext_lazy("Update password service URL"),
                "help_text": ugettext_lazy(
                    "The URL of an external page where users will be able"
                    " to update their password. It applies only to non local"
                    " users, ie. those automatically created after a successful"
                    " external authentication (LDAP, SMTP)."
                )
            }),
            ("password_recovery_msg", {
                "label": ugettext_lazy("Password recovery announcement"),
                "help_text": ugettext_lazy(
                    "A temporary message that will be displayed on the "
                    "reset password page."
                )
            }),
            ("sms_password_recovery", {
                "label": ugettext_lazy("Enable password recovery by SMS"),
                "help_text": ugettext_lazy(
                    "Enable password recovery by SMS for users who filled "
                    "a phone number."
                )
            }),
            ("sms_provider", {
                "label": ugettext_lazy("SMS provider"),
                "display": "sms_password_recovery=true",
                "help_text": ugettext_lazy(
                    "Choose a provider to send password recovery SMS"
                )
            }),
            *sms_backends.get_all_backend_structures()
        ])
    }),
    ("ldap", {
        "label": _("LDAP"),
        "params": collections.OrderedDict([
            ("ldap_server_address", {
                "label": ugettext_lazy("Server address"),
                "help_text": ugettext_lazy(
                    "The IP address or the DNS name of the LDAP server"),
            }),
            ("ldap_server_port", {
                "label": ugettext_lazy("Server port"),
                "help_text": ugettext_lazy(
                    "The TCP port number used by the LDAP server")
            }),
            ("ldap_enable_secondary_server", {
                "label": ugettext_lazy("Enable secondary server (fallback)"),
                "help_text": ugettext_lazy(
                    "Enable a secondary LDAP server which will be used "
                    "if the primary one fails"
                )
            }),
            ("ldap_secondary_server_address", {
                "label": ugettext_lazy("Secondary server address"),
                "display": "ldap_enable_secondary_server=true",
                "help_text": ugettext_lazy(
                    "The IP address or the DNS name of the seondary LDAP server"
                )
            }),
            ("ldap_secondary_server_port", {
                "label": ugettext_lazy("Secondary server port"),
                "display": "ldap_enable_secondary_server=true",
                "help_text": ugettext_lazy(
                    "The TCP port number used by the LDAP secondary server"
                )
            }),
            ("ldap_secured", {
                "label": ugettext_lazy("Use a secured connection"),
                "help_text": ugettext_lazy(
                    "Use an SSL/STARTTLS connection to access the LDAP server")
            }),
            ("ldap_is_active_directory", {
                "label": ugettext_lazy("Active Directory"),
                "help_text": ugettext_lazy(
                    "Tell if the LDAP server is an Active Directory one")
            }),
            ("ldap_admin_groups", {
                "label": ugettext_lazy("Administrator groups"),
                "help_text": ugettext_lazy(
                    "Members of those LDAP Posix groups will be created as "
                    "domain administrators. Use ';' characters to separate "
                    "groups."
                )
            }),
            ("ldap_group_type", {
                "label": ugettext_lazy("Group type"),
                "help_text": ugettext_lazy(
                    "The LDAP group type to use with your directory."
                )
            }),
            ("ldap_groups_search_base", {
                "label": ugettext_lazy("Groups search base"),
                "help_text": ugettext_lazy(
                    "The distinguished name of the search base used to find "
                    "groups"
                )
            }),
            ("ldap_password_attribute", {
                "label": ugettext_lazy("Password attribute"),
                "help_text": ugettext_lazy(
                    "The attribute used to store user passwords"),
            }),
            ("ldap_auth_sep", {
                "label": ugettext_lazy("LDAP authentication settings"),
                "display": "authentication_type=ldap",
                "separator": True
            }),
            ("ldap_auth_method", {
                "label": ugettext_lazy("Authentication method"),
                "display": "authentication_type=ldap",
                "help_text": ugettext_lazy(
                    "Choose the authentication method to use"),
            }),
            ("ldap_bind_dn", {
                "label": ugettext_lazy("Bind DN"),
                "help_text": ugettext_lazy(
                    "The distinguished name to use when binding to the LDAP "
                    "server. Leave empty for an anonymous bind"
                ),
                "display": "authentication_type=ldap&ldap_auth_method=searchbind"
            }),
            ("ldap_bind_password", {
                "label": ugettext_lazy("Bind password"),
                "help_text": ugettext_lazy(
                    "The password to use when binding to the LDAP server "
                    "(with 'Bind DN')"
                ),
                "display": "authentication_type=ldap&ldap_auth_method=searchbind",
                "password": True
            }),
            ("ldap_search_base", {
                "label": ugettext_lazy("Users search base"),
                "help_text": ugettext_lazy(
                    "The distinguished name of the search base used to find "
                    "users"
                ),
                "display": "authentication_type=ldap&ldap_auth_method=searchbind"
            }),
            ("ldap_search_filter", {
                "label": ugettext_lazy("Search filter"),
                "help_text": ugettext_lazy(
                    "An optional filter string (e.g. '(objectClass=person)'). "
                    "In order to be valid, it must be enclosed in parentheses."
                ),
                "display": "authentication_type=ldap&ldap_auth_method=searchbind"
            }),
            ("ldap_user_dn_template", {
                "label": ugettext_lazy("User DN template"),
                "help_text": ugettext_lazy(
                    "The template used to construct a user's DN. It should "
                    "contain one placeholder (ie. %(user)s)"
                ),
                "display": "authentication_type=ldap&ldap_auth_method=directbind"
            }),
            ("ldap_sync_sep", {
                "label": ugettext_lazy("LDAP synchronization settings"),
                "separator": True
            }),
            ("ldap_sync_bind_dn", {
                "label": ugettext_lazy("Bind DN"),
                "help_text": ugettext_lazy(
                    "The distinguished name to use when binding to the LDAP server. "
                    "Leave empty for an anonymous bind"
                )
            }),
            ("ldap_sync_bind_password", {
                "label": ugettext_lazy("Bind password"),
                "help_text": ugettext_lazy(
                    "The password to use when binding to the LDAP server "
                    "(with 'Bind DN')"
                ),
                "password": True
            }),
            ("ldap_enable_sync", {
                "label": ugettext_lazy("Enable export to LDAP"),
                "help_text": ugettext_lazy(
                    "Enable automatic synchronization between local database "
                    "and LDAP directory"
                )
            }),
            ("ldap_sync_delete_remote_account", {
                "label": ugettext_lazy(
                    "Delete remote LDAP account when local account is deleted"),
                "help_text": ugettext_lazy(
                    "Delete remote LDAP account when local account is deleted, "
                    "otherwise it will be disabled."
                ),
                "display": "ldap_enable_sync=true"
            }),
            ("ldap_sync_account_dn_template", {
                "label": ugettext_lazy("Account DN template"),
                "help_text": ugettext_lazy(
                    "The template used to construct an account's DN. It should "
                    "contain one placeholder (ie. %(user)s)"
                ),
                "display": "ldap_enable_sync=true"
            }),
            ("ldap_enable_import", {
                "label": ugettext_lazy("Enable import from LDAP"),
                "help_text": ugettext_lazy(
                    "Enable account synchronization from LDAP directory to "
                    "local database"
                )
            }),
            ("ldap_import_search_base", {
                "label": ugettext_lazy("Users search base"),
                "help_text": ugettext_lazy(
                    "The distinguished name of the search base used to find "
                    "users"
                ),
                "display": "ldap_enable_import=true"
            }),
            ("ldap_import_search_filter", {
                "label": ugettext_lazy("Search filter"),
                "help_text": ugettext_lazy(
                    "An optional filter string (e.g. '(objectClass=person)'). "
                    "In order to be valid, it must be enclosed in parentheses."
                ),
                "display": "ldap_enable_import=true"
            }),
            ("ldap_import_username_attr", {
                "label": ugettext_lazy("Username attribute"),
                "help_text": ugettext_lazy(
                    "The name of the LDAP attribute where the username can be "
                    "found."
                ),
                "display": "ldap_enable_import=true"
            }),
            ("ldap_dovecot_sync", {
                "label": ugettext_lazy("Enable Dovecot LDAP sync"),
                "help_text": ugettext_lazy(
                    "LDAP authentication settings will be applied to Dovecot "
                    "configuration."
                )
            }),
            ("ldap_dovecot_conf_file", {
                "label": ugettext_lazy("Dovecot LDAP config file"),
                "help_text": ugettext_lazy(
                    "Location of the configuration file which contains "
                    "Dovecot LDAP settings."
                ),
                "display": "ldap_dovecot_sync=true"
            }),

        ])
    }),
    ("dashboard", {
        "label": ugettext_lazy("Dashboard"),
        "params": collections.OrderedDict([
            ("rss_feed_url", {
                "label": ugettext_lazy("Custom RSS feed"),
                "help_text": ugettext_lazy(
                    "Display custom RSS feed to resellers and domain "
                    "administrators"
                )
            }),
            ("hide_features_widget", {
                "label": ugettext_lazy("Hide features widget"),
                "help_text": ugettext_lazy(
                    "Hide features widget for resellers and domain "
                    "administrators"
                )
            }),
        ])
    }),
    ("notifications", {
        "label": ugettext_lazy("Notifications"),
        "params": collections.OrderedDict([
            ("sender_address", {
                "label": ugettext_lazy("Sender address"),
                "help_text": ugettext_lazy(
                    "Email address used to send notifications.")
            })
        ])
    }),
    ("api", {
        "label": ugettext_lazy("Public API"),
        "params": collections.OrderedDict([
            ("enable_api_communication", {
                "label": ugettext_lazy("Enable communication"),
                "help_text": ugettext_lazy(
                    "Automatically checks if a newer version is available")
            }),
            ("check_new_versions", {
                "label": ugettext_lazy("Check new versions"),
                "display": "enable_api_communication=true",
                "help_text": ugettext_lazy(
                    "Automatically checks if a newer version is available")
            }),
            ("send_new_versions_email", {
                "label": ugettext_lazy(
                    "Send an email when new versions are found"),
                "display": "check_new_versions=true",
                "help_text": ugettext_lazy(
                    "Send an email to notify admins about new versions"
                )
            }),
            ("new_versions_email_rcpt", {
                "label": ugettext_lazy("Recipient"),
                "display": "check_new_versions=true",
                "help_text": ugettext_lazy(
                    "Recipient of new versions notification emails."
                )
            }),
            ("send_statistics", {
                "label": ugettext_lazy("Send statistics"),
                "display": "enable_api_communication=true",
                "help_text": ugettext_lazy(
                    "Send statistics to Modoboa public API "
                    "(counters and used extensions)")
            })
        ])
    }),
    ("misc", {
        "label": ugettext_lazy("Miscellaneous"),
        "params": collections.OrderedDict([
            ("inactive_account_threshold", {
                "label": ugettext_lazy("Inactive account threshold"),
                "help_text": ugettext_lazy(
                    "An account with a last login date greater than this "
                    "threshold (in days) will be considered as inactive"
                )
            }),
            ("top_notifications_check_interval", {
                "label": ugettext_lazy("Top notifications check interval"),
                "help_text": ugettext_lazy(
                    "Interval between two top notification checks (in seconds)"
                ),
            }),
            ("log_maximum_age", {
                "label": ugettext_lazy("Maximum log record age"),
                "help_text": ugettext_lazy(
                    "The maximum age in days of a log record"),
            }),
            ("items_per_page", {
                "label": ugettext_lazy("Items per page"),
                "help_text": ugettext_lazy("Number of displayed items per page")
            }),
            ("default_top_redirection", {
                "label": ugettext_lazy("Default top redirection"),
                "help_text": ugettext_lazy(
                    "The default redirection used when no application is "
                    "specified"
                ),
            })
        ])
    })
])



"""Core API serializers."""

from django.utils.translation import ugettext as _

from rest_framework import serializers


class CheckTFASetupSerializer(serializers.Serializer):
    """Serializer used to finalize 2FA setup."""

    pin_code = serializers.CharField()

    def validate_pin_code(self, value):
        device = self.context["user"].totpdevice_set.first()
        if not device.verify_token(value):
            raise serializers.ValidationError(_("Invalid PIN code"))
        return value

    def validate(self, data):
        user = self.context["user"]
        if user.tfa_enabled or user.staticdevice_set.exists():
            raise serializers.ValidationError(_("2FA is already enabled"))
        return data



# Generated by Django 2.2.9 on 2020-04-21 06:51

from django.db import migrations
import phonenumber_field.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0019_auto_20200121_1659'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='phone_number',
            field=phonenumber_field.modelfields.PhoneNumberField(blank=True, max_length=128, null=True, region=None, verbose_name='Phone number'),
        ),
    ]



"""Parameters serializers."""

from rest_framework import serializers


class ApplicationSerializer(serializers.Serializer):
    """Application serializer."""

    name = serializers.CharField()
    label = serializers.CharField()


class ParameterChoiceSerializer(serializers.Serializer):
    """Parameter choice serializer."""

    text = serializers.CharField()
    value = serializers.CharField()


class ParameterSerializer(serializers.Serializer):
    """Parameter serializer."""

    name = serializers.CharField()
    label = serializers.CharField()
    help_text = serializers.CharField()
    display = serializers.CharField()
    widget = serializers.CharField()
    choices = ParameterChoiceSerializer(many=True)


class StructureSerializer(serializers.Serializer):
    """Structure serializer."""

    name = serializers.CharField()
    display = serializers.CharField(help_text="Display rule")
    parameters = ParameterSerializer(many=True)



from functools import reduce

from ckeditor_uploader import views as cku_views

from django.conf import settings
from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.urls import include, path
from django.views.i18n import JavaScriptCatalog

from drf_spectacular.views import (
    SpectacularJSONAPIView, SpectacularSwaggerView, SpectacularRedocView
)

from modoboa.admin.views import user as user_views
from modoboa.core import signals as core_signals, views as core_views
from modoboa.core.extensions import exts_pool

urlpatterns = [
    path('jsi18n/', JavaScriptCatalog.as_view(), name="javascript-catalog"),

    path('ckeditor/upload/', login_required(cku_views.upload),
         name="ckeditor_upload"),
    path('ckeditor/browse/', login_required(cku_views.browse),
         name="ckeditor_browse"),

    path('', include("modoboa.core.urls")),
    path('admin/', include("modoboa.admin.urls")),
    path('dnstools/', include("modoboa.dnstools.urls")),
    path('stats/', include("modoboa.maillog.urls")),

    path('user/forward/', user_views.forward, name="user_forward"),

    path('accounts/password_reset/', core_views.PasswordResetView.as_view(),
         name="password_reset"),
    path('accounts/password_reset/done/',
         auth_views.PasswordResetDoneView.as_view(),
         name="password_reset_done"),
    path('reset/confirm_code/',
         core_views.VerifySMSCodeView.as_view(),
         name="password_reset_confirm_code"),
    path('reset/resend_code/',
         core_views.ResendSMSCodeView.as_view(),
         name="password_reset_resend_code"),
    path('reset/<uidb64>/<token>/',
         auth_views.PasswordResetConfirmView.as_view(),
         name="password_reset_confirm"),
    path('reset/done/', auth_views.PasswordResetCompleteView.as_view(),
         name="password_reset_complete"),
]

exts_pool.load_all()
urlpatterns += exts_pool.get_urls()

extra_routes = core_signals.extra_uprefs_routes.send(sender="urls")
if extra_routes:
    extra_routes = reduce(
        lambda a, b: a + b, [route[1] for route in extra_routes])
    urlpatterns += extra_routes

# API urls
urlpatterns += [
    # FIXME: legacy, to remove ASAP
    path('docs/openapi.json',
         SpectacularJSONAPIView.as_view(api_version="v1"), name='schema-v1-legacy'),

    path('api/schema-v1/',
         SpectacularJSONAPIView.as_view(api_version="v1"), name='schema-v1'),
    path('api/schema-v1/swagger/',
         SpectacularSwaggerView.as_view(url_name='schema-v1'),
         name='docs-index-v1'),
    path('api/schema-v1/redoc/',
         SpectacularRedocView.as_view(url_name='schema-v1')),

    path('api/schema-v2/',
         SpectacularJSONAPIView.as_view(api_version="v2"), name='schema-v2'),
    path('api/schema-v2/swagger/',
         SpectacularSwaggerView.as_view(url_name='schema-v2'),
         name='docs-index-v2'),
    path('api/schema-v2/redoc/',
         SpectacularRedocView.as_view(url_name='schema-v2')),
    path('api/v1/', include("modoboa.urls_api_v1", namespace="v1")),
    path('api/v2/', include("modoboa.urls_api_v2", namespace="v2")),
]

if settings.DEBUG:
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns
    from django.conf.urls.static import static
    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT
    )



"""Export related forms."""

from django import forms
from django.utils.translation import ugettext_lazy


class ExportDataForm(forms.Form):
    sepchar = forms.CharField(
        label=ugettext_lazy("Separator"),
        max_length=1,
        required=False,
        widget=forms.TextInput(attrs={"class": "form-control"})
    )

    def __init__(self, *args, **kwargs):
        super(ExportDataForm, self).__init__(*args, **kwargs)
        self.fields["sepchar"].widget.attrs = {"class": "col-md-1 form-control"}

    def clean_sepchar(self):
        if self.cleaned_data["sepchar"] == "":
            return ";"
        return self.cleaned_data["sepchar"]



"""
A simple managemenent command to update the default site.

See `https://docs.djangoproject.com/en/dev/ref/contrib/sites/`_.
"""

from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):

    """Management command to set the default site."""

    help = "Set default site (see django.contrib.sites)"  # NOQA:A003

    def add_arguments(self, parser):
        """Define command arguments."""
        parser.add_argument("hostname", type=str)

    def handle(self, *args, **options):
        """Command entry point."""
        if "hostname" not in options:
            raise CommandError("You must provide a hostname")
        site = Site.objects.get(pk=1)
        site.domain = options["hostname"]
        site.name = options["hostname"]
        site.save()



"""Django management command to import admin objects."""

import csv
import io
import os

import progressbar
from chardet.universaldetector import UniversalDetector

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import ugettext as _

from modoboa.core import models as core_models
from modoboa.core.extensions import exts_pool
from modoboa.lib.exceptions import Conflict
from .... import signals


class ImportCommand(BaseCommand):
    """Command class."""

    args = "csvfile"
    help = "Import identities from a csv file"  # NOQA:A003

    def add_arguments(self, parser):
        """Add extra arguments to command."""
        parser.add_argument(
            "--sepchar", type=str, default=";",
            help="Separator used in file.")
        parser.add_argument(
            "--continue-if-exists", action="store_true",
            dest="continue_if_exists", default=False,
            help="Continue even if an entry already exists.")
        parser.add_argument(
            "--crypt-password", action="store_true",
            default=False, help="Encrypt provided passwords.")
        parser.add_argument(
            "files", type=str, nargs="+", help="CSV files to import.")

    def _import(self, filename, options, encoding="utf-8"):
        """Import domains or identities."""
        superadmin = (
            core_models.User.objects.filter(is_superuser=True).first()
        )
        if not os.path.isfile(filename):
            raise CommandError("File not found")

        num_lines = sum(
            1 for line in io.open(filename, encoding=encoding) if line
        )
        pbar = progressbar.ProgressBar(
            widgets=[
                progressbar.Percentage(), progressbar.Bar(), progressbar.ETA()
            ], maxval=num_lines
        ).start()
        with io.open(filename, encoding=encoding, newline="") as f:
            reader = csv.reader(f, delimiter=options["sepchar"])
            i = 0
            for row in reader:
                if not row:
                    continue
                fct = signals.import_object.send(
                    sender=self.__class__, objtype=row[0].strip())
                fct = [func for x_, func in fct if func is not None]
                if not fct:
                    continue
                fct = fct[0]
                try:
                    fct(superadmin, row, options)
                except Conflict:
                    if options["continue_if_exists"]:
                        continue
                    raise CommandError(
                        "Object already exists: {}".format(
                            options["sepchar"].join(row[:2])))
                i += 1
                pbar.update(i)

        pbar.finish()

    def handle(self, *args, **options):
        """Command entry point."""
        exts_pool.load_all()
        for filename in options["files"]:
            try:
                with transaction.atomic():
                    self._import(filename, options)
            except CommandError as exc:
                raise exc
            except UnicodeDecodeError:
                self.stdout.write(self.style.NOTICE(
                    _("CSV file is not encoded in UTF-8, attempting to guess "
                      "encoding")
                ))
                detector = UniversalDetector()
                with io.open(filename, "rb") as fp:
                    for line in fp:
                        detector.feed(line)
                        if detector.done:
                            break
                    detector.close()

                self.stdout.write(self.style.NOTICE(
                    _("Reading CSV file using %(encoding)s encoding") %
                    detector.result
                ))
                try:
                    with transaction.atomic():
                        self._import(
                            filename, options,
                            encoding=detector.result["encoding"]
                        )
                except UnicodeDecodeError as exc:
                    raise CommandError(
                        _("Unable to decode CSV file using %(encoding)s "
                          "encoding") % detector.result
                    ) from exc



"""A shortcut to deploy a fresh modoboa instance."""

import getpass
import os
from os.path import isfile
import shutil
import subprocess
import sys

import dj_database_url

import django
from django.core import management
from django.template import Context, Template
from django.utils.encoding import smart_str

from modoboa.core.commands import Command
from modoboa.lib.api_client import ModoAPIClient
from modoboa.lib.sysutils import exec_cmd

DBCONN_TPL = """
    '{{ conn_name }}': {
        'ENGINE': '{{ ENGINE }}',
        'NAME': '{{ NAME }}',
        'USER': '{% if USER %}{{ USER }}{% endif %}',
        'PASSWORD': '{% if PASSWORD %}{{ PASSWORD }}{% endif %}',
        'HOST': '{% if HOST %}{{ HOST }}{% endif %}',
        'PORT': '{% if PORT %}{{ PORT }}{% endif %}',
        'ATOMIC_REQUESTS': True,
        {% if ENGINE == 'django.db.backends.mysql' %}'OPTIONS' : {
            "init_command" : 'SET foreign_key_checks = 0;',
        },{% endif %}
    },
"""


class DeployCommand(Command):

    """The ``deploy`` command."""

    help = (  # NOQA:A003
        "Create a fresh django project (calling startproject)"
        " and apply Modoboa specific settings."
    )

    def __init__(self, *args, **kwargs):
        super(DeployCommand, self).__init__(*args, **kwargs)
        self._parser.add_argument("name", type=str,
                                  help="The name of your Modoboa instance")
        self._parser.add_argument(
            "--collectstatic", action="store_true", default=False,
            help="Run django collectstatic command"
        )
        self._parser.add_argument(
            "--dburl", type=str, nargs="+", default=None,
            help="A database-url with a name")
        self._parser.add_argument(
            "--domain", type=str, default=None,
            help="The domain under which you want to deploy modoboa")
        self._parser.add_argument(
            "--lang", type=str, default="en",
            help="Set the default language"
        )
        self._parser.add_argument(
            "--timezone", type=str, default="UTC",
            help="Set the local timezone"
        )
        self._parser.add_argument(
            "--devel", action="store_true", default=False,
            help="Create a development instance"
        )
        self._parser.add_argument(
            "--extensions", type=str, nargs="*",
            help="The list of extension to deploy"
        )
        self._parser.add_argument(
            "--dont-install-extensions", action="store_true", default=False,
            help="Do not install extensions using pip"
        )
        self._parser.add_argument(
            "--admin-username", default="admin",
            help="Username of the initial super administrator"
        )

    def _exec_django_command(self, name, cwd, *args):
        """Run a django command for the freshly created project

        :param name: the command name
        :param cwd: the directory where the command must be executed
        """
        cmd = [sys.executable, "manage.py", name]
        cmd.extend(args)
        if not self._verbose:
            p = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd
            )
            output = p.communicate()
        else:
            p = subprocess.Popen(cmd, cwd=cwd)
            p.wait()
            output = None
        if p.returncode:
            if output:
                print(
                    "\n".join([l.decode() for l in output if l is not None]),
                    file=sys.stderr
                )
            print("%s failed, check your configuration" % cmd, file=sys.stderr)

    def ask_db_info(self, name="default"):
        """Prompt the user for database information

        Gather all information required to create a new database
        connection (into settings.py).

        :param name: the connection name
        """
        print("Configuring database connection: %s" % name)
        info = {
            "conn_name": name,
            "ENGINE": input(
                "Database type (mysql, postgres or sqlite3): ")
        }
        if info["ENGINE"] not in ["mysql", "postgres", "sqlite3"]:
            raise RuntimeError("Unsupported database engine")

        if info["ENGINE"] == "sqlite3":
            info["ENGINE"] = "django.db.backends.sqlite3"
            info["NAME"] = "%s.db" % name
            return info
        if info["ENGINE"] == "postgres":
            info["ENGINE"] = "django.db.backends.postgresql"
            default_port = 5432
        else:
            info["ENGINE"] = "django.db.backends.mysql"
            default_port = 3306
        info["HOST"] = input("Database host (default: 'localhost'): ")
        info["PORT"] = input(
            "Database port (default: '%s'): " % default_port)
        # leave port setting empty, if default value is supplied and
        # leave it to django
        if info["PORT"] == default_port:
            info["PORT"] = ""
        info["NAME"] = input("Database name: ")
        info["USER"] = input("Username: ")
        info["PASSWORD"] = getpass.getpass("Password: ")
        return info

    def _get_extension_list(self):
        """Ask the API to get the list of all extensions.

        We hardcode the API url here to avoid a loading of
        django's settings since they are not available yet...
        """
        url = "http://api.modoboa.org/"
        official_exts = ModoAPIClient(url).list_extensions()
        return [extension["name"] for extension in official_exts]

    def find_extra_settings(self, extensions):
        """Install one or more extensions.

        Return the list of extensions providing settings we must
        include in the final configuration.

        """
        extra_settings = []
        for extension in extensions:
            module = __import__(extension[1], locals(), globals(), [])
            basedir = os.path.dirname(module.__file__)
            if not os.path.exists("{0}/settings.py".format(basedir)):
                continue
            extra_settings.append(extension[1])
        return extra_settings

    def handle(self, parsed_args):
        django.setup()
        management.call_command(
            "startproject", parsed_args.name, verbosity=False
        )
        path = "%(name)s/%(name)s" % {"name": parsed_args.name}
        sys.path.append(parsed_args.name)

        conn_tpl = Template(DBCONN_TPL)
        connections = {}
        if parsed_args.dburl:
            for dburl in parsed_args.dburl:
                conn_name, url = dburl.split(":", 1)
                info = dj_database_url.config(default=url)
                # In case the user fails to supply a valid database url,
                # fallback to manual mode
                if not info:
                    print("There was a problem with your database-url. \n")
                    info = self.ask_db_info(conn_name)
                # If we set this earlier, our fallback method will never
                # be triggered
                info["conn_name"] = conn_name
                connections[conn_name] = conn_tpl.render(Context(info))
        else:
            connections["default"] = conn_tpl.render(
                Context(self.ask_db_info()))

        if parsed_args.domain:
            allowed_host = parsed_args.domain
        else:
            allowed_host = input(
                "What will be the hostname used to access Modoboa? ")
            if not allowed_host:
                allowed_host = "localhost"
        extra_settings = []
        extensions = parsed_args.extensions
        if extensions:
            if "all" in extensions:
                extensions = self._get_extension_list()
            extensions = [(extension, extension.replace("-", "_"))
                          for extension in extensions]
            if not parsed_args.dont_install_extensions:
                cmd = (
                    sys.executable +
                    " -m pip install " +
                    " ".join([extension[0] for extension in extensions])
                )
                exec_cmd(cmd, capture_output=False)
            extra_settings = self.find_extra_settings(extensions)
            extensions = [extension[1] for extension in extensions]

        bower_components_dir = os.path.realpath(
            os.path.join(os.path.dirname(__file__), "../../bower_components")
        )

        mod = __import__(
            parsed_args.name, globals(), locals(), [smart_str("settings")]
        )
        tpl = self._render_template(
            "%s/settings.py.tpl" % self._templates_dir, {
                "db_connections": connections,
                "secret_key": mod.settings.SECRET_KEY,
                "name": parsed_args.name,
                "allowed_host": allowed_host,
                "lang": parsed_args.lang,
                "timezone": parsed_args.timezone,
                "bower_components_dir": bower_components_dir,
                "devmode": parsed_args.devel,
                "extensions": extensions,
                "extra_settings": extra_settings
            }
        )
        with open("%s/settings.py" % path, "w") as fp:
            fp.write(tpl)
        shutil.copyfile(
            "%s/urls.py.tpl" % self._templates_dir, "%s/urls.py" % path
        )
        os.mkdir("%s/media" % parsed_args.name)

        if isfile("%s/settings.pyc" % path):
            os.unlink("%s/settings.pyc" % path)
        self._exec_django_command(
            "migrate", parsed_args.name, "--noinput"
        )
        self._exec_django_command(
            "load_initial_data", parsed_args.name,
            "--admin-username", parsed_args.admin_username
        )
        if parsed_args.collectstatic:
            self._exec_django_command(
                "collectstatic", parsed_args.name, "--noinput"
            )
        self._exec_django_command(
            "set_default_site", parsed_args.name, allowed_host
        )

        base_frontend_dir = os.path.join(
            os.path.dirname(__file__), "../../frontend_dist/")
        if not os.path.exists(base_frontend_dir):
            return
        frontend_target_dir = "{}/frontend".format(parsed_args.name)
        shutil.copytree(base_frontend_dir, frontend_target_dir)

        with open("{}/config.json".format(frontend_target_dir), "w") as fp:
            fp.write("""{
    "API_BASE_URL": "https://%s/api/v2"
}
""" % allowed_host)



"""Forms related to aliases management."""

from collections import OrderedDict

from django import forms
from django.http import QueryDict
from django.utils.translation import ugettext as _, ugettext_lazy

from modoboa.core import signals as core_signals
from modoboa.lib import exceptions as lib_exceptions, fields as lib_fields
from modoboa.lib.email_utils import split_mailbox
from modoboa.lib.form_utils import DynamicForm
from ..models import Alias, Domain


class AliasForm(forms.ModelForm, DynamicForm):
    """A form to create/modify an alias."""

    random_address = forms.BooleanField(
        label=ugettext_lazy("Random address"),
        required=False)
    address = lib_fields.UTF8AndEmptyUserEmailField(
        label=ugettext_lazy("Email address"),
        help_text=ugettext_lazy(
            "The alias address. To create a catchall alias, just enter the "
            "domain name (@domain.tld)."
        ),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )
    recipients = lib_fields.UTF8AndEmptyUserEmailField(
        label=ugettext_lazy("Recipients"), required=False,
        help_text=ugettext_lazy(
            "Addresses this alias will point to. Indicate only one address "
            "per input, press ENTER to add a new input."
        ),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )

    class Meta:
        model = Alias
        fields = ("address", "domain", "enabled", "expire_at", "description")
        labels = {
            "domain": ugettext_lazy("Domain")
        }
        widgets = {
            "domain": forms.widgets.Select(
                attrs={"class": "selectpicker", "data-live-search": "true"})
        }

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(AliasForm, self).__init__(*args, **kwargs)
        self.fields = OrderedDict(
            (key, self.fields[key]) for key in
            ["random_address", "address", "domain", "recipients", "enabled",
             "expire_at", "description"]
        )
        if self.instance.pk:
            del self.fields["random_address"]
            del self.fields["domain"]
        else:
            self.fields["address"].required = False
            self.fields["domain"].required = False
            self.fields["domain"].queryset = Domain.objects.get_for_admin(user)
        if len(args) and isinstance(args[0], QueryDict):
            if "instance" in kwargs:
                if not kwargs["instance"].domain.enabled:
                    del self.fields["enabled"]
            self._load_from_qdict(args[0], "recipients", forms.EmailField)
        elif "instance" in kwargs:
            alias = kwargs["instance"]
            if not alias.domain.enabled:
                self.fields["enabled"].widget.attrs["disabled"] = "disabled"
            cpt = 1
            for rcpt in alias.aliasrecipient_set.filter(alias__internal=False):
                name = "recipients_%d" % cpt
                self._create_field(forms.EmailField, name, rcpt.address, 2)
                cpt += 1

    def clean_address(self):
        """Check if address points to a local domain."""
        if not self.cleaned_data["address"]:
            return self.cleaned_data["address"]
        localpart, domname = split_mailbox(self.cleaned_data["address"])
        try:
            domain = Domain.objects.get(name=domname)
        except Domain.DoesNotExist:
            raise forms.ValidationError(_("Domain does not exist"))
        if not self.user.can_access(domain):
            raise forms.ValidationError(
                _("You don't have access to this domain")
            )
        if not self.instance.pk:
            try:
                core_signals.can_create_object.send(
                    sender=self.__class__, context=domain,
                    object_type="mailbox_aliases")
            except lib_exceptions.ModoboaException as inst:
                raise forms.ValidationError(inst)
        return self.cleaned_data["address"].lower()

    def clean(self):
        """Check it there is at least one recipient."""
        super(AliasForm, self).clean()
        random_address = self.cleaned_data.get("random_address")
        if not random_address:
            if not self.cleaned_data.get("address"):
                self.add_error("address", _("This field is required"))
        elif not self.cleaned_data.get("domain"):
            self.add_error("domain", _("This field is required"))
        for field, value in self.cleaned_data.items():
            if field.startswith("recipients") and value:
                return self.cleaned_data
        self.add_error("recipients", _("No recipient defined"))
        return self.cleaned_data

    def save(self, commit=True):
        """Custom save method."""
        alias = super(AliasForm, self).save(commit=False)
        if self.cleaned_data.get("random_address"):
            alias.address = "{}@{}".format(
                Alias.generate_random_address(), alias.domain)
        else:
            local_part, domname = split_mailbox(self.cleaned_data["address"])
            alias.domain = Domain.objects.get(name=domname)
        if commit:
            alias.save()
            address_list = [
                value for field, value in self.cleaned_data.items()
                if field.startswith("recipients") and value
            ]
            alias.set_recipients(address_list)
        return alias



# Generated by Django 1.10.7 on 2017-12-20 14:17
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('transport', '0001_initial'),
        ('admin', '0010_auto_20170215_1328'),
    ]

    operations = [
        migrations.AddField(
            model_name='domain',
            name='transport',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, to='transport.Transport'),
        ),
    ]



from django.core import exceptions
from django.core.files.base import ContentFile
from django.urls import reverse

from modoboa.admin import factories as admin_factories, models as admin_models
from modoboa.core import factories as core_factories, models as core_models
from modoboa.lib.tests import ModoTestCase
from .. import utils


class LimitImportTestCase(ModoTestCase):
    """Base class to test limits."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(LimitImportTestCase, cls).setUpTestData()
        admin_factories.populate_database()
        cls.reseller = core_factories.UserFactory(
            username="reseller", groups=("Resellers", ))
        cls.dadmin = core_models.User.objects.get(username="admin@test.com")
        cls.domain = admin_models.Domain.objects.get(name="test.com")

    def _test_domain_alias_import(self, limit):
        """Check domain aliases limit."""
        self.client.force_login(self.reseller)
        self.assertFalse(limit.is_exceeded())
        f = ContentFile("""domainalias; domalias1.com; {domain}; True
domainalias; domalias2.com; {domain}; True
""".format(domain=self.domain), name="domains.csv")
        self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        self.assertTrue(limit.is_exceeded())

        f = ContentFile(
            "domainalias; domalias3.com; {}; True".format(self.domain),
            name="domains.csv")
        response = self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        self.assertContains(response, "Domain aliases: limit reached")

    def _test_domain_admins_import(self, limit, initial_count=0):
        """Check domain admins limit."""
        self.client.force_login(self.reseller)
        self.assertFalse(limit.is_exceeded())
        content = ""
        for cpt in range(initial_count, 2):
            content += (
                "account; admin{cpt}@{domain}; toto; User; One; True; "
                "DomainAdmins; user{cpt}@{domain}; 5; {domain}\n"
                .format(domain=self.domain, cpt=cpt)
            )
        f = ContentFile(content, name="domain_admins.csv")
        response = self.client.post(
            reverse("admin:identity_import"), {
                "sourcefile": f
            }
        )
        self.assertTrue(limit.is_exceeded())

        f = ContentFile("""account; admin3@{domain}; toto; User; One; True; DomainAdmins; admin3@{domain}; 5; {domain}
""".format(domain=self.domain), name="domain_admins.csv")  # NOQA:E501
        response = self.client.post(
            reverse("admin:identity_import"), {
                "sourcefile": f
            }
        )
        self.assertContains(response, "Permission denied")

    def _test_mailboxes_import(self, limit):
        """Check mailboxes limit."""
        self.client.force_login(self.dadmin)
        self.assertFalse(limit.is_exceeded())
        f = ContentFile(
            "account; user1@{domain}; toto; User; One; True; SimpleUsers; "
            "user1@{domain}; 5\r\n"
            "account; truc@{domain}; toto; René; Truc; True; SimpleUsers; "
            "truc@{domain}; 5\r\n".format(domain=self.domain),
            name="domains.csv"
        )
        response = self.client.post(
            reverse("admin:identity_import"), {
                "sourcefile": f
            }
        )
        self.assertTrue(limit.is_exceeded())

        f = ContentFile("""
account; user3@{domain}; toto; User; One; True; SimpleUsers; user3@{domain}; 5
""".format(domain=self.domain), name="domains.csv")
        response = self.client.post(
            reverse("admin:identity_import"), {
                "sourcefile": f
            }
        )
        self.assertContains(response, "Mailboxes: limit reached")

    def _test_mailbox_aliases_import(self, limit):
        """Check mailbox aliases limit."""
        self.client.force_login(self.dadmin)
        self.assertFalse(limit.is_exceeded())
        f = ContentFile("""alias; alias1@{domain}; True; user@{domain}
alias; alias2@{domain}; True; user@{domain}
""".format(domain=self.domain), name="aliases.csv")
        self.client.post(
            reverse("admin:identity_import"), {
                "sourcefile": f
            }
        )
        self.assertTrue(limit.is_exceeded())

        f = ContentFile("""
alias; alias3@{domain}; True; user@{domain}
""".format(domain=self.domain), name="aliases.csv")
        with self.assertRaises(exceptions.ValidationError):
            self.client.post(
                reverse("admin:identity_import"), {
                    "sourcefile": f
                }
            )


class UserLimitImportTestCase(LimitImportTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        localconfig = core_models.LocalConfig.objects.first()
        for name, _definition in utils.get_user_limit_templates():
            localconfig.parameters.set_value(
                "deflt_user_{0}_limit".format(name), 2)
        localconfig.save()
        super(UserLimitImportTestCase, cls).setUpTestData()

    def test_domains_import(self):
        """Check domains limit."""
        self.client.force_login(self.reseller)
        limit = self.reseller.userobjectlimit_set.get(name="domains")
        self.assertFalse(limit.is_exceeded())
        f = ContentFile("domain; domain1.com; 1; 1; True", name="domains.csv")
        self.client.post(
            reverse("admin:domain_import"), {"sourcefile": f}
        )
        self.assertFalse(limit.is_exceeded())

        f = ContentFile("domain; domain2.com; 0; 1; True", name="domains.csv")
        response = self.client.post(
            reverse("admin:domain_import"), {"sourcefile": f}
        )
        self.assertContains(
            response, "not allowed to define unlimited values")

        f = ContentFile("domain; domain2.com; 2; 1; True", name="domains.csv")
        response = self.client.post(
            reverse("admin:domain_import"), {"sourcefile": f}
        )
        self.assertContains(response, "Quota: limit reached")

        f = ContentFile(b"""domain; domain2.com; 1; 1; True
domain; domain3.com; 1000; 100; True
""", name="domains.csv")
        response = self.client.post(
            reverse("admin:domain_import"), {"sourcefile": f}
        )
        self.assertContains(response, "Domains: limit reached")

    def test_domain_alias_import(self):
        """Check domain aliases limit."""
        limit = self.reseller.userobjectlimit_set.get(name="domain_aliases")
        self._test_domain_alias_import(limit)

    def test_domain_admins_import(self):
        """Check domain admins limit."""
        self.domain.add_admin(self.reseller)
        limit = self.reseller.userobjectlimit_set.get(name="domain_admins")
        self._test_domain_admins_import(limit)

    def test_mailboxes_import(self):
        """Check mailboxes limit."""
        limit = self.dadmin.userobjectlimit_set.get(name="mailboxes")
        self._test_mailboxes_import(limit)

    def test_mailbox_aliases_import(self):
        """Check mailbox aliases limit."""
        limit = self.dadmin.userobjectlimit_set.get(name="mailbox_aliases")
        self._test_mailbox_aliases_import(limit)


class DomainLimitImportTestCase(LimitImportTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        localconfig = core_models.LocalConfig.objects.first()
        localconfig.parameters.set_values({
            "enable_domain_limits": True,
            "enable_admin_limits": False
        })
        for name, _definition in utils.get_domain_limit_templates():
            localconfig.parameters.set_value(
                "deflt_domain_{0}_limit".format(name), 2)
        localconfig.save()
        super(DomainLimitImportTestCase, cls).setUpTestData()
        mb = admin_factories.MailboxFactory(
            user__username="user@test4.com",
            domain__name="test4.com", address="user")
        cls.domain = mb.domain

    def test_domain_alias_import(self):
        """Check domain aliases limit."""
        limit = self.domain.domainobjectlimit_set.get(name="domain_aliases")
        self._test_domain_alias_import(limit)

    def test_mailboxes_import(self):
        """Check mailboxes limit."""
        self.domain.add_admin(self.dadmin)
        limit = self.domain.domainobjectlimit_set.get(name="mailboxes")
        self._test_mailboxes_import(limit)

    def test_mailbox_aliases_import(self):
        """Check mailbox aliases limit."""
        self.domain.add_admin(self.dadmin)
        limit = self.domain.domainobjectlimit_set.get(name="mailbox_aliases")
        self._test_mailbox_aliases_import(limit)



"""AppConfig for relaydomains."""

from django.apps import AppConfig


class RelayDomainsConfig(AppConfig):
    """App configuration."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "modoboa.relaydomains"
    verbose_name = "Modoboa relay domains"

    def ready(self):
        from . import handlers  # NOQA:F401
        from . import transport  # NOQA:F401



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('limits', '0003_auto_20160413_1046'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='limit',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='limit',
            name='pool',
        ),
        migrations.RemoveField(
            model_name='limitspool',
            name='user',
        ),
        migrations.DeleteModel(
            name='Limit',
        ),
        migrations.DeleteModel(
            name='LimitsPool',
        ),
    ]



# Generated by Django 2.2.12 on 2020-09-16 11:46

from django.db import migrations


def rename_settings(apps, schema_editor):
    LocalConfig = apps.get_model("core", "LocalConfig")
    lc = LocalConfig.objects.first()
    if not lc:
        return
    stats_params = lc._parameters.pop("modoboa_stats", None)
    if stats_params:
        lc._parameters["maillog"] = stats_params
        lc.save()


def backwards(apps, schema_editor):
    LocalConfig = apps.get_model("core", "LocalConfig")
    lc = LocalConfig.objects.first()
    if not lc:
        return
    params = lc._parameters.pop("maillog", None)
    if params:
        lc._parameters["modoboa_stats"] = params
        lc.save()


class Migration(migrations.Migration):

    dependencies = [
        ('maillog', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(rename_settings, backwards),
    ]



# Generated by Django 2.2.12 on 2020-12-04 08:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0017_alarm'),
    ]

    operations = [
        migrations.AlterField(
            model_name='alarm',
            name='title',
            field=models.TextField(),
        ),
    ]



from django.db import models, migrations

APPLICATIONS = [
    ("admin", "modoboa_admin"),
    ("amavis", "modoboa_amavis"),
    ("limits", "modoboa_admin_limits"),
    ("postfix_autoreply", "modoboa_postfix_autoreply"),
    ("postfix_relay_domains", "modoboa_admin_relaydomains"),
    ("radicale", "modoboa_radicale"),
    ("stats", "modoboa_stats"),
    ("sievefilters", "modoboa_sievefilters"),
    ("webmail", "modoboa_webmail"),
]

def rename_app_parameters(app, model):
    """Rename all parameters for a given app."""
    qset = model.objects.filter(name__startswith=app[0])
    for param in qset:
        param.name = param.name.replace(
            "{}.".format(app[0]), "{}.".format(app[1])
        )
        param.save()


def rename_parameters(apps, schema_editor):
    """Rename old parameters."""
    Parameter = apps.get_model("lib", "Parameter")
    UserParameter = apps.get_model("lib", "UserParameter")
    for app in APPLICATIONS:
        rename_app_parameters(app, Parameter)
        rename_app_parameters(app, UserParameter)


class Migration(migrations.Migration):

    dependencies = [
        ('lib', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(rename_parameters)
    ]



"""Forms related to forwards management."""

from django import forms
from django.core import validators
from django.utils.translation import ugettext_lazy


class ForwardForm(forms.Form):
    """Forward definition form."""

    dest = forms.CharField(
        label=ugettext_lazy("Recipient(s)"),
        widget=forms.Textarea(attrs={"class": "form-control"}),
        required=False,
        help_text=ugettext_lazy(
            "Indicate one or more recipients separated by a ','")
    )
    keepcopies = forms.BooleanField(
        label=ugettext_lazy("Keep local copies"),
        required=False,
        help_text=ugettext_lazy(
            "Forward messages and store copies into your local mailbox")
    )

    def clean_dest(self):
        """Check recipients validity."""
        rawdata = self.cleaned_data.get("dest", "").strip()
        recipients = []
        if not rawdata:
            return recipients
        for rcpt in rawdata.split(","):
            validators.validate_email(rcpt)
            recipients += [rcpt]
        return recipients



"""Forms related to import operations."""

from django import forms
from django.utils.translation import ugettext_lazy


class ImportDataForm(forms.Form):

    """Base form to import objects."""

    sourcefile = forms.FileField(label=ugettext_lazy("Select a file"))
    sepchar = forms.CharField(
        label=ugettext_lazy("Separator"),
        max_length=1,
        required=False,
        widget=forms.TextInput(attrs={"class": "form-control"})
    )
    continue_if_exists = forms.BooleanField(
        label=ugettext_lazy("Continue on error"), required=False,
        help_text=ugettext_lazy("Don't treat duplicated objects as error")
    )

    def __init__(self, *args, **kwargs):
        super(ImportDataForm, self).__init__(*args, **kwargs)
        self.fields["sepchar"].widget.attrs = {"class": "col-md-1 form-control"}

    def clean_sepchar(self):
        if self.cleaned_data["sepchar"] == "":
            return ";"
        return self.cleaned_data["sepchar"]


class ImportIdentitiesForm(ImportDataForm):

    """A form to import identities."""

    crypt_password = forms.BooleanField(
        label=ugettext_lazy("Crypt passwords"), required=False,
        help_text=ugettext_lazy(
            "Check this option if passwords contained in your file "
            "are not crypted"
        )
    )



"""User related tests."""

from django.urls import reverse

from modoboa.core.models import User
from modoboa.lib.tests import ModoTestCase
from ..factories import populate_database
from ..models import Alias


class ForwardTestCase(ModoTestCase):
    """User forward test cases."""

    def setUp(self):
        super(ForwardTestCase, self).setUp()
        populate_database()

    def test_set_forward(self):
        self.client.logout()
        self.client.login(username="user@test.com", password="toto")
        url = reverse("user_forward")
        self.ajax_post(
            url,
            {"dest": "user@extdomain.com", "keepcopies": True}
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        content = response.json()
        self.assertIn("user@extdomain.com", content["content"])
        self.assertNotIn("user@test.com", content["content"])
        forward = Alias.objects.get(address="user@test.com", internal=False)
        sadmin = User.objects.get(username="admin")
        self.assertTrue(sadmin.can_access(forward))
        domadmin = User.objects.get(username="admin@test.com")
        self.assertTrue(domadmin.can_access(forward))

        selfalias = Alias.objects.get(address="user@test.com", internal=True)
        self.assertTrue(selfalias.enabled)

        # Deactivate local copies
        self.ajax_post(
            url,
            {"dest": "user@extdomain.com", "keepcopies": False}
        )
        selfalias.refresh_from_db()
        self.assertFalse(selfalias.enabled)

        # Now, empty forward
        self.ajax_post(url, {"dest": "", "keepcopies": False})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        content = response.json()
        self.assertNotIn("user@extdomain.com", content["content"])
        selfalias.refresh_from_db()
        self.assertTrue(selfalias.enabled)



"""Tests for web_utils."""

from django.test import SimpleTestCase

from .. import web_utils


class TestCase(SimpleTestCase):
    """Test functions."""

    def test_size2integer(self):
        self.assertEqual(web_utils.size2integer("1024"), 1024)
        # Convert to bytes
        self.assertEqual(web_utils.size2integer("1K"), 1024)
        self.assertEqual(web_utils.size2integer("1M"), 1048576)
        self.assertEqual(web_utils.size2integer("1G"), 1073741824)
        # Convert to megabytes
        self.assertEqual(web_utils.size2integer("1K", output_unit="MB"), 0)
        self.assertEqual(web_utils.size2integer("1M", output_unit="MB"), 1)
        self.assertEqual(web_utils.size2integer("1G", output_unit="MB"), 1024)
        # Unsupported unit
        with self.assertRaises(ValueError):
            web_utils.size2integer("1K", output_unit="GB")



"""Custom renderers for DRF."""

import csv
from io import StringIO

from rest_framework import renderers


class CSVRenderer(renderers.BaseRenderer):
    """Custom CSV renderer."""

    media_type = "text/csv"
    format = "csv"

    def render(self, data, media_type=None, renderer_context=None):
        with StringIO() as fp:
            csvwriter = csv.writer(fp)
            for item in data:
                csvwriter.writerow(item)
            content = fp.getvalue()
        return content



"""Custom signals."""

from threading import local

_local_store = local()


def set_current_request(request):
    """Store current request."""
    _local_store.request = request


def get_request():
    """Get the current request from anywhere."""
    if "request" not in _local_store.__dict__:
        return None
    return _local_store.request



"""App related API urls."""

from rest_framework import routers

from . import viewsets

router = routers.SimpleRouter()
router.register(
    r"statistics", viewsets.StatisticsViewSet, basename="statistics")
router.register(
    r"logs/messages", viewsets.MaillogViewSet, basename="maillog")

urlpatterns = router.urls



"""Repair command tests"""

from django.core import management

from modoboa.lib.permissions import ObjectAccess, get_object_owner
from modoboa.lib.tests import ModoTestCase
from .. import factories, models


class RepairTestCase(ModoTestCase):
    """TestCase for repair command."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create some data."""
        super(RepairTestCase, cls).setUpTestData()
        factories.populate_database()

    def test_management_command(self):
        """Check that command works fine."""
        ObjectAccess.objects.all().delete()
        mbox = models.Mailbox.objects.first()
        alias = models.Alias.objects.first()
        # assert mbox has no owner
        self.assertIs(get_object_owner(mbox), None)
        # fix it. run in quiet mode because we dont want output in tests
        ret = management.call_command("modo", "repair", "--quiet")
        assert ret is None
        # assert it's fixed
        self.assertIsNot(get_object_owner(mbox), None)
        self.assertIsNot(get_object_owner(alias), None)

    def test_management_command_with_dry_run(self):
        """Check that command works fine."""
        ObjectAccess.objects.all().delete()
        mbox = models.Mailbox.objects.first()
        # assert mbox has no owner
        self.assertIs(get_object_owner(mbox), None)
        # show problems. run in quiet mode because we dont want output in tests
        ret = management.call_command("modo", "repair", "--quiet", "--dry-run")
        assert ret is None
        # assert its not fixed
        self.assertIs(get_object_owner(mbox), None)

    def test_management_command_with_nul_domain(self):
        """Just assume nothing raise when an alias has no domain."""
        models.Alias.objects.create(address="@modoboa.xxx")
        ret = management.call_command("modo", "repair", "--quiet")
        assert ret is None

    def test_management_command_with_no_alias(self):
        """Check that problem is fixed."""
        count, detail = models.Alias.objects.filter(
            address="user@test.com", internal=True).delete()
        self.assertEqual(count, 3)
        ret = management.call_command("modo", "repair", "--quiet")
        assert ret is None
        self.assertTrue(
            models.Alias.objects.filter(
                address="user@test.com", internal=True).exists())



"""API v2 tests."""

from django.urls import reverse

from modoboa.admin import factories as admin_factories
from modoboa.admin import models as admin_models
from modoboa.dnstools import factories
from modoboa.lib.tests import ModoAPITestCase


class DNSViewSetTestCase(ModoAPITestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super().setUpTestData()
        admin_factories.populate_database()
        cls.spf_rec = factories.DNSRecordFactory(
            type="spf", value="v=SPF1 mx -all", is_valid=True,
            domain__name="test.com"
        )
        cls.dmarc_rec = factories.DNSRecordFactory(
            type="dmarc", value="XXX", is_valid=False,
            error="Not a DMARC record",
            domain__name="test.com"
        )
        cls.dkim_rec = factories.DNSRecordFactory(
            type="dkim", value="12345", is_valid=False,
            error="Public key mismatchs",
            domain__name="test.com"
        )
        cls.ac_rec = factories.DNSRecordFactory(
            type="autoconfig", value="1.2.3.4", is_valid=True,
            domain__name="test.com"
        )

    def test_dns_detail(self):
        domain = admin_models.Domain.objects.get(name="test.com")
        url = reverse("v2:dns-dns-detail", args=[domain.pk])
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()["dmarc_record"]["type"], "dmarc")



"""Models related to domains management."""

import datetime

from reversion import revisions as reversion

from django.db import models
from django.utils import timezone
from django.utils.encoding import force_text, smart_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext as _, ugettext_lazy

from modoboa.core import signals as core_signals
from modoboa.core.models import User
from modoboa.lib.exceptions import BadRequest, Conflict
from modoboa.parameters import tools as param_tools

from .. import constants
from .base import AdminObject
from . import mixins


class Domain(mixins.MessageLimitMixin, AdminObject):
    """Mail domain."""

    name = models.CharField(ugettext_lazy("name"), max_length=253, unique=True,
                            help_text=ugettext_lazy("The domain name"))
    quota = models.PositiveIntegerField(
        default=0,
    )
    default_mailbox_quota = models.PositiveIntegerField(
        verbose_name=ugettext_lazy("Default mailbox quota"),
        default=0
    )
    message_limit = models.PositiveIntegerField(
        ugettext_lazy("Message sending limit"),
        null=True, blank=True,
        help_text=ugettext_lazy(
            "Number of messages this domain can send per day"
        )
    )
    enabled = models.BooleanField(
        ugettext_lazy("enabled"),
        help_text=ugettext_lazy("Check to activate this domain"),
        default=True
    )
    type = models.CharField(default="domain", max_length=20)  # NOQA:A003
    enable_dns_checks = models.BooleanField(
        ugettext_lazy("Enable DNS checks"), default=True,
        help_text=ugettext_lazy("Check to enable DNS checks for this domain")
    )

    transport = models.OneToOneField(
        "transport.Transport", null=True, on_delete=models.SET_NULL)

    enable_dkim = models.BooleanField(
        ugettext_lazy("Enable DKIM signing"),
        help_text=ugettext_lazy(
            "If you activate this feature, a DKIM key will be "
            "generated for this domain."),
        default=False
    )
    dkim_key_selector = models.CharField(max_length=30, default="modoboa")
    dkim_key_length = models.PositiveIntegerField(
        ugettext_lazy("Key length"), choices=constants.DKIM_KEY_LENGTHS,
        blank=True, null=True
    )
    dkim_public_key = models.TextField(blank=True)
    dkim_private_key_path = models.CharField(max_length=254, blank=True)

    class Meta:
        ordering = ["name"]
        app_label = "admin"

    def __init__(self, *args, **kwargs):
        """Save name for further use."""
        super(Domain, self).__init__(*args, **kwargs)
        self.old_mail_homes = None
        self.oldname = self.name
        self.old_dkim_key_length = self.dkim_key_length

    @property
    def domainalias_count(self) -> int:
        return self.domainalias_set.count()

    @property
    def mailbox_count(self) -> int:
        return self.mailbox_set.count()

    @property
    def mbalias_count(self) -> int:
        return self.alias_set.filter(internal=False).count()

    @property
    def identities_count(self):
        """Total number of identities in this domain."""
        return (
            self.mailbox_set.count() +
            self.alias_set.filter(internal=False).count())

    @property
    def tags(self):
        label = ""
        for dt in constants.DOMAIN_TYPES:
            if self.type == dt[0]:
                label = dt[1]
        result = [{"name": self.type, "label": label, "type": "dom"}]
        if self.transport:
            result.append({
                "name": self.transport.service,
                "label": self.transport.service,
                "type": "srv",
                "color": "info"
            })
        return result

    @property
    def admins(self):
        """Return the domain administrators of this domain.

        :return: a list of User objects
        """
        return User.objects.filter(
            is_superuser=False,
            objectaccess__content_type__model="domain",
            objectaccess__object_id=self.pk)

    @property
    def aliases(self):
        return self.domainalias_set

    @property
    def uses_a_reserved_tld(self):
        """Does this domain use a reserved TLD."""
        tld = self.name.split(".", 1)[-1]
        return tld in constants.RESERVED_TLD

    @property
    def just_created(self):
        """Return true if the domain was created in the latest 24h."""
        now = timezone.now()
        delta = datetime.timedelta(days=1)
        return self.creation + delta > now

    def awaiting_checks(self):
        """Return true if the domain has no valid MX record and was created
        in the latest 24h."""
        if (not self.mxrecord_set.has_valids()) and self.just_created:
            return True
        return False

    @property
    def dns_global_status(self) -> str:
        """Return global DNS status."""
        if not self.enable_dns_checks or self.uses_a_reserved_tld:
            return "disabled"
        elif self.awaiting_checks():
            return "pending"
        config = dict(param_tools.get_global_parameters("admin"))
        errors = []
        if config["enable_mx_checks"] and not self.mxrecord_set.has_valids():
            errors.append("mx")
        if config["enable_dnsbl_checks"] and self.dnsblresult_set.blacklisted().exists():
            errors.append("dnsbl")
        if config["enable_spf_checks"]:
            if self.spf_record is None or not self.spf_record.is_valid:
                errors.append("spf")
            if self.dkim_record is None or not self.dkim_record.is_valid:
                errors.append("dkim")
            if self.dmarc_record is None or not self.dmarc_record.is_valid:
                errors.append("dmarc")
        if config["enable_autoconfig_checks"]:
            if self.autoconfig_record is None:
                errors.append("autoconfig")
            if self.autodiscover_record is None:
                errors.append("autodiscover")
        if len(errors) == 0:
            return "ok"
        return "critical"

    @cached_property
    def dnsbl_status_color(self):
        """Shortcut to DNSBL results."""
        if not self.dnsblresult_set.exists():
            return "warning"
        elif self.dnsblresult_set.blacklisted().exists():
            return "danger"
        else:
            return "success"

    @property
    def spf_record(self):
        """Return SPF record."""
        return self.dnsrecord_set.filter(type="spf").first()

    @property
    def dkim_record(self):
        """Return DKIM record."""
        return self.dnsrecord_set.filter(type="dkim").first()

    @property
    def dmarc_record(self):
        """Return DMARC record."""
        return self.dnsrecord_set.filter(type="dmarc").first()

    @property
    def autoconfig_record(self):
        """Return autoconfig record."""
        return self.dnsrecord_set.filter(type="autoconfig").first()

    @property
    def autodiscover_record(self):
        """Return autodiscover record."""
        return self.dnsrecord_set.filter(type="autodiscover").first()

    @cached_property
    def allocated_quota(self):
        """Return current quota allocation."""
        if not self.quota:
            return 0
        if not self.mailbox_set.exists():
            return 0
        return self.mailbox_set.aggregate(
            total=models.Sum("quota"))["total"]

    @cached_property
    def allocated_quota_in_percent(self):
        """Return allocated quota in percent."""
        if not self.allocated_quota:
            return 0
        return int(self.allocated_quota / float(self.quota) * 100)

    @cached_property
    def used_quota(self):
        """Return current quota usage."""
        from .mailbox import Quota

        if not self.quota:
            return 0
        if not self.mailbox_set.exists():
            return 0
        return int(Quota.objects.get_domain_usage(self) / 1048576)

    @cached_property
    def used_quota_in_percent(self):
        """Return used quota in percent."""
        if not self.allocated_quota:
            return 0
        return int(self.used_quota / float(self.quota) * 100)

    @property
    def message_counter_key(self):
        """Return the key used to store messages count."""
        return self.name

    @property
    def bind_format_dkim_public_key(self):
        """Return DKIM public key using Bind format."""
        record = "v=DKIM1;k=rsa;p=%s" % self.dkim_public_key
        # TXT records longer than 255 characters need split into chunks
        # split record into 74 character chunks (+2 for indent, +2 for quotes(")
        # == 78 characters) to make more readable in text editors.
        split_record = []
        while record:
            if len(record) > 74:
                split_record.append("  \"%s\"" % record[:74])
                record = record[74:]
            else:
                split_record.append("  \"%s\"" % record)
                break
        record = "\n".join(split_record)
        return "{}._domainkey.{}. IN TXT (\n{})".format(
            self.dkim_key_selector, self.name, record)

    def add_admin(self, account):
        """Add a new administrator to this domain.

        :param User account: the administrator
        """
        from modoboa.lib.permissions import grant_access_to_object

        core_signals.can_create_object.send(
            sender=self.__class__, context=self, object_type="domain_admins")
        grant_access_to_object(account, self)
        for mb in self.mailbox_set.all():
            if mb.user.has_perm("admin.add_domain"):
                continue
            grant_access_to_object(account, mb)
            grant_access_to_object(account, mb.user)
        for al in self.alias_set.all():
            grant_access_to_object(account, al)

    def remove_admin(self, account):
        """Remove an administrator of this domain.

        :param User account: administrator to remove
        """
        from modoboa.lib.permissions import ungrant_access_to_object

        ungrant_access_to_object(self, account)
        for mb in self.mailbox_set.all():
            if mb.user.has_perm("admin.add_domain"):
                continue
            ungrant_access_to_object(mb, account)
            ungrant_access_to_object(mb.user, account)
        for al in self.alias_set.all():
            ungrant_access_to_object(al, account)

    def save(self, *args, **kwargs):
        """Store current data if domain is renamed."""
        if self.oldname != self.name:
            self.old_mail_homes = {
                mb.id: mb.mail_home for mb in self.mailbox_set.all()
            }
        if self.old_dkim_key_length != self.dkim_key_length:
            self.dkim_public_key = ""
            self.dkim_private_key_path = ""
        super(Domain, self).save(*args, **kwargs)

    def delete(self, fromuser, keepdir=False):
        """Custom delete method."""
        from modoboa.lib.permissions import ungrant_access_to_objects
        from .mailbox import Quota

        if self.domainalias_set.count():
            ungrant_access_to_objects(self.domainalias_set.all())
        if self.alias_set.count():
            ungrant_access_to_objects(self.alias_set.all())
        if param_tools.get_global_parameter("auto_account_removal"):
            User.objects.filter(mailbox__domain=self).delete()
        elif self.mailbox_set.count():
            Quota.objects.filter(username__contains="@%s" % self.name).delete()
            ungrant_access_to_objects(self.mailbox_set.all())
        super(Domain, self).delete()

    def __str__(self):
        return smart_text(self.name)

    def from_csv(self, user, row):
        """Create a new domain from a CSV entry.

        The expected fields order is the following::

          "domain", name, quota, default_mailbox_quota, enabled

        :param ``core.User`` user: user creating the domain
        :param str row: a list containing domain's definition
        """
        from .. import lib

        if len(row) < 5:
            raise BadRequest(_("Invalid line"))
        self.name = row[1].strip().lower()
        if Domain.objects.filter(name=self.name).exists():
            raise Conflict
        domains_must_have_authorized_mx = (
            param_tools.get_global_parameter("domains_must_have_authorized_mx")
        )
        if domains_must_have_authorized_mx and not user.is_superuser:
            if not lib.domain_has_authorized_mx(self.name):
                raise BadRequest(
                    _("No authorized MX record found for domain {}")
                    .format(self.name)
                )
        try:
            self.quota = int(row[2].strip())
        except ValueError:
            raise BadRequest(
                _("Invalid quota value for domain '{}'")
                .format(self.name)
            )
        try:
            self.default_mailbox_quota = int(row[3].strip())
        except ValueError:
            raise BadRequest(
                _("Invalid default mailbox quota value for domain '{}'")
                .format(self.name)
            )
        if self.quota != 0 and self.default_mailbox_quota > self.quota:
            raise BadRequest(
                _("Default mailbox quota cannot be greater than domain "
                  "quota")
            )
        self.enabled = (row[4].strip().lower() in ["true", "1", "yes", "y"])
        core_signals.can_create_object.send(
            sender=self.__class__, context=user, klass=Domain,
            instance=self)
        self.save(creator=user)

    def to_csv_rows(self):
        """Return a row to include in a CSV file."""
        result = [
            [
                "domain",
                force_text(self.name),
                self.quota,
                self.default_mailbox_quota,
                self.enabled
            ]
        ]
        for dalias in self.domainalias_set.all():
            result.append(dalias.to_csv_row())
        return result

    def to_csv(self, csvwriter):
        """Export domain and domain aliases to CSV format."""
        for row in self.to_csv_rows():
            csvwriter.writerow(row)

    def post_create(self, creator):
        """Post creation actions.

        :param ``User`` creator: user whos created this domain
        """
        super().post_create(creator)
        for domalias in self.domainalias_set.all():
            domalias.post_create(creator)


reversion.register(Domain)



"""App related constants."""

import re


DNS_RECORD_TYPES = [
    ("spf", "SPF"),
    ("dkim", "DKIM"),
    ("dmarc", "DMARC"),
    ("autoconfig", "Autoconfig"),
    ("autodiscover", "Autodiscover"),
]

SPF_MECHANISMS = ["ip4", "ip6", "a", "mx", "ptr", "exists", "include"]

DMARC_URI_REGEX = re.compile(r"^mailto:(.+)(!\w+)?")

DMARC_TAGS = {
    "adkim": {"values": ["r", "s"]},
    "aspf": {"values": ["r", "s"]},
    "fo": {"values": ["0", "1", "d", "s"]},
    "p": {"values": ["none", "quarantine", "reject"]},
    "pct": {"type": "int", "min_value": 0, "max_value": 100},
    "rf": {"type": "list", "values": ["afrf"]},
    "ri": {"type": "int"},
    "rua": {"type": "list", "regex": DMARC_URI_REGEX},
    "ruf": {"type": "list", "regex": DMARC_URI_REGEX},
    "sp": {"values": ["none", "quarantine", "reject"]}
}



"""Field validators."""

import re

from django.core.exceptions import ValidationError
from django.core.validators import EmailValidator, URLValidator
from django.utils.translation import ugettext_lazy


class HostnameValidator(object):

    """Validator for fqdn."""

    message = ugettext_lazy("Enter a valid domain name")
    code = "invalid"
    regex = re.compile(URLValidator.host_re)

    def __init__(self, message=None, code=None):
        """Constructor."""
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        """Check value."""
        if len(value) > 255:
            raise ValidationError(self.message, self.code)
        if value[-1] == ".":
            # strip exactly one dot from the right, if present.
            value = value[:-1]
        if not self.regex.match(value):
            raise ValidationError(self.message, self.code)


validate_hostname = HostnameValidator()


class UTF8EmailValidator(EmailValidator):

    """Validator for addresses using non-ASCII characters."""

    # unicode letters range (must be a unicode string, not a raw string)
    ul = "\u00a1-\uffff"
    ascii_set = "-!#$%&'*+/=?^_`{}|~0-9A-Z"
    user_regex_raw = (
        # dot-atom
        r"^[" + ascii_set + ul + r"]+(\.[" + ascii_set + ul + r"]+)*\Z"
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|'
        r'\\[\001-\011\013\014\016-\177])*"\Z)'
    )
    user_regex = re.compile(r"(" + user_regex_raw, re.IGNORECASE)


validate_utf8_email = UTF8EmailValidator()


class UTF8AndEmptyUserEmailValidator(UTF8EmailValidator):

    """Same as upper + allows empty local part."""

    user_regex = re.compile(
        r"(^$|" + UTF8EmailValidator.user_regex_raw, re.IGNORECASE)


validate_utf8_and_empty_user_email = UTF8AndEmptyUserEmailValidator()



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_user_language'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='phone_number',
            field=models.CharField(max_length=128, null=True, verbose_name='Phone number', blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='user',
            name='secondary_email',
            field=models.EmailField(help_text='An alternative e-mail address, can be used for recovery needs.', max_length=254, null=True, verbose_name='Secondary email', blank=True),
            preserve_default=True,
        ),
    ]



# Generated by Django 1.11.8 on 2018-01-12 15:58
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0011_domain_transport'),
    ]

    operations = [
        migrations.AddField(
            model_name='domain',
            name='dkim_key_length',
            field=models.PositiveIntegerField(blank=True, choices=[(1024, '1024'), (2048, '2048'), (4096, '4096')], null=True, verbose_name='Key length'),
        ),
        migrations.AddField(
            model_name='domain',
            name='dkim_key_selector',
            field=models.CharField(default='modoboa', max_length=30, verbose_name='Key selector'),
        ),
        migrations.AddField(
            model_name='domain',
            name='dkim_private_key_path',
            field=models.CharField(blank=True, max_length=254),
        ),
        migrations.AddField(
            model_name='domain',
            name='dkim_public_key',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='domain',
            name='enable_dkim',
            field=models.BooleanField(default=False, help_text='If you activate this feature, a DKIM key will be generated for this domain.', verbose_name='Enable DKIM signing'),
        ),
    ]



"""Tests for u2u_decode."""

from django.test import TestCase

from .. import u2u_decode


class U2UTestCase(TestCase):
    """Test RFC1342 decoding utilities."""

    def test_header_decoding(self):
        """Simple decoding."""
        samples = [
            ("=?ISO-8859-15?Q?=20Profitez de tous les services en ligne sur "
             "impots.gouv.fr?=",
             "Profitez de tous les services en ligne sur impots.gouv.fr"),
            ("=?ISO-8859-1?Q?Accus=E9?= de =?ISO-8859-1?Q?r=E9ception?= de "
             "votre annonce",
             "Accusé de réception de votre annonce"),
            ("Sm=?ISO-8859-1?B?9g==?=rg=?ISO-8859-1?B?5Q==?=sbord",
             "Sm\xf6rg\xe5sbord"),
            # The following case currently fails because of the way we split
            # encoded words to parse them separately, which can lead to
            # unexpected unicode decode errors... I think it will work fine on
            # Python3
            # ("=?utf-8?B?VMOpbMOpcMOpYWdlIFZJTkNJIEF1dG9yb3V0ZXMgLSBFeHDD?=\n"
            #  "=?utf-8?B?qWRpdGlvbiBkZSB2b3RyZSBjb21tYW5kZSBOwrAgMjAxNzEyMDcw"
            #  "MDA1?=\n=?utf-8?B?MyBkdSAwNy8xMi8yMDE3IDE0OjQ5OjQx?=",
            #  "")
        ]
        for sample in samples:
            self.assertEqual(u2u_decode.u2u_decode(sample[0]), sample[1])

    def test_address_header_decoding(self):
        """Check address decoding."""
        mailsploit_sample = (
            "=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?==?utf-8?Q?=00?="
            "=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?=@mailsploit.com")
        expected_result = (
            "=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?==?utf-8?Q??="
            "=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?=@mailsploit.com")
        self.assertEqual(
            u2u_decode.decode_address(mailsploit_sample),
            ("", expected_result)
        )
        mailsploit_sample = (
            '"=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?==?utf-8?Q?=0A=00?="\n'
            "<=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?==?utf-8?Q?=0A=00?="
            "@mailsploit.com>")
        expected_result = (
            "potus@whitehouse.gov",
            "=?utf-8?b?cG90dXNAd2hpdGVob3VzZS5nb3Y=?==?utf-8?Q??="
            "@mailsploit.com")
        self.assertEqual(
            u2u_decode.decode_address(mailsploit_sample),
            expected_result)



# Generated by Django 1.10.7 on 2017-11-24 11:16
from django.db import migrations, models

import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Transport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pattern', models.CharField(max_length=254, unique=True, verbose_name='pattern')),
                ('service', models.CharField(max_length=30, verbose_name='service')),
                ('next_hop', models.CharField(blank=True, max_length=100, verbose_name='next hop')),
                ('_settings', jsonfield.fields.JSONField(default={})),
            ],
            options={
                'ordering': ['pattern'],
            },
        ),
    ]



# Generated by Django 1.9.5 on 2016-11-04 18:15
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0006_auto_20161104_1900'),
        ('relaydomains', '0005_auto_20161105_1426'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='alias',
            name='dates',
        ),
        migrations.RemoveField(
            model_name='domain',
            name='dates',
        ),
        migrations.RemoveField(
            model_name='domainalias',
            name='dates',
        ),
        migrations.RemoveField(
            model_name='mailbox',
            name='dates',
        ),
        migrations.AlterField(
            model_name='alias',
            name='last_modification',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='domain',
            name='last_modification',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='domainalias',
            name='last_modification',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='mailbox',
            name='last_modification',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.DeleteModel(
            name='ObjectDates',
        ),
    ]



"""App. related models."""

from django.db import models


class Maillog(models.Model):
    """A model to store message logs."""

    queue_id = models.CharField(max_length=50)
    date = models.DateTimeField(db_index=True)
    sender = models.EmailField()
    rcpt = models.EmailField()
    original_rcpt = models.EmailField(null=True)
    size = models.PositiveIntegerField()
    status = models.CharField(max_length=15)

    from_domain = models.ForeignKey(
        "admin.Domain", on_delete=models.SET_NULL, null=True,
        related_name="sent_messages_log"
    )
    to_domain = models.ForeignKey(
        "admin.Domain", on_delete=models.SET_NULL, null=True,
        related_name="recv_messages_log"
    )

    class Meta:
        ordering = ['date']



"""Admin serializers."""

from typing import List

from django.conf import settings
from django.contrib.auth import password_validation
from django.core.exceptions import ValidationError
from django.shortcuts import get_object_or_404
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _, ugettext_lazy

from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from modoboa.admin import models as admin_models
from modoboa.core import (
    constants as core_constants, models as core_models, signals as core_signals
)
from modoboa.lib import (
    email_utils, exceptions as lib_exceptions, fields as lib_fields,
    permissions, validators, web_utils
)
from modoboa.parameters import tools as param_tools

from ... import lib, models


class DomainSerializer(serializers.ModelSerializer):
    """Base Domain serializer."""

    quota = serializers.CharField(
        required=False,
        help_text=ugettext_lazy(
            "Quota shared between mailboxes. Can be expressed in KB, "
            "MB (default) or GB. A value of 0 means no quota."
        )
    )
    default_mailbox_quota = serializers.CharField(
        required=False,
        help_text=ugettext_lazy(
            "Default quota in MB applied to mailboxes. A value of 0 means "
            "no quota."
        )
    )

    class Meta:
        model = models.Domain
        fields = (
            "pk", "name", "quota", "default_mailbox_quota", "enabled", "type",
            "enable_dkim", "dkim_key_selector", "dkim_key_length",
            "dkim_public_key", "dkim_private_key_path",
            "mailbox_count", "mbalias_count", "domainalias_count",
            "dns_global_status", "message_limit", "creation",
            "last_modification"
        )
        read_only_fields = (
            "pk", "dkim_public_key", "dns_global_status", "allocated_quota_in_percent"
            "mailbox_count", "mbalias_count", "domainalias_count",
            "enable_dns_checks", "creation", "last_modification"
        )

    def validate_name(self, value):
        """Check name constraints."""
        domains_must_have_authorized_mx = (
            param_tools.get_global_parameter("domains_must_have_authorized_mx")
        )
        user = self.context["request"].user
        value = value.lower()
        if domains_must_have_authorized_mx and not user.is_superuser:
            if not lib.domain_has_authorized_mx(value):
                raise serializers.ValidationError(
                    _("No authorized MX record found for this domain"))
        return value

    def validate_quota(self, value):
        """Convert quota to MB."""
        return web_utils.size2integer(value, output_unit="MB")

    def validate_default_mailbox_quota(self, value):
        """Convert quota to MB."""
        return web_utils.size2integer(value, output_unit="MB")

    def validate(self, data):
        """Check quota values."""
        quota = data.get("quota", 0)
        default_mailbox_quota = data.get("default_mailbox_quota", 0)
        if quota != 0 and default_mailbox_quota > quota:
            raise serializers.ValidationError({
                "default_mailbox_quota":
                _("Cannot be greater than domain quota")
            })
        return data

    def create(self, validated_data):
        """Set permissions."""
        params = dict(param_tools.get_global_parameters("admin"))
        domain = models.Domain(**validated_data)
        condition = (
            params["default_domain_message_limit"] is not None and
            "message_limit" not in validated_data
        )
        if condition:
            domain.message_limit = params["default_domain_message_limit"]
        creator = self.context["request"].user
        core_signals.can_create_object.send(
            sender=self.__class__, context=creator,
            klass=models.Domain, instance=domain)
        domain.save(creator=creator)
        return domain


class DomainAliasSerializer(serializers.ModelSerializer):
    """Base DomainAlias serializer."""

    class Meta:
        model = admin_models.DomainAlias
        fields = ("pk", "name", "target", "enabled", )
        expandable_fields = {
            "target": DomainSerializer
        }

    def validate_target(self, value):
        """Check target domain."""
        if not self.context["request"].user.can_access(value):
            raise serializers.ValidationError(_("Permission denied."))
        return value

    def validate_name(self, value):
        """Lower case name."""
        return value.lower()

    def create(self, validated_data):
        """Custom creation."""
        domain_alias = models.DomainAlias(**validated_data)
        creator = self.context["request"].user
        try:
            core_signals.can_create_object.send(
                sender=self.__class__, context=creator,
                klass=models.DomainAlias)
            core_signals.can_create_object.send(
                sender=self.__class__, context=domain_alias.target,
                object_type="domain_aliases")
        except lib_exceptions.ModoboaException as inst:
            raise serializers.ValidationError({
                "domain": force_text(inst)})
        domain_alias.save(creator=creator)
        return domain_alias


class MailboxSerializer(serializers.ModelSerializer):
    """Base mailbox serializer."""

    full_address = lib_fields.DRFEmailFieldUTF8()
    quota = serializers.CharField(required=False)

    class Meta:
        model = models.Mailbox
        fields = (
            "pk", "full_address", "use_domain_quota", "quota", "message_limit"
        )

    def validate_full_address(self, value):
        """Lower case address."""
        return value.lower()

    def validate_quota(self, value):
        """Convert quota to MB."""
        return web_utils.size2integer(value, output_unit="MB")

    def validate(self, data):
        """Check if quota is required."""
        method = self.context["request"].method
        if not data.get("use_domain_quota", False):
            if "quota" not in data and method != "PATCH":
                raise serializers.ValidationError({
                    "quota": _("This field is required")
                })
        return data


class AccountSerializer(serializers.ModelSerializer):
    """Base account serializer."""

    role = serializers.SerializerMethodField()
    mailbox = MailboxSerializer(required=False)
    domains = serializers.SerializerMethodField(
        help_text=_(
            "List of administered domains (resellers and domain "
            "administrators only)."))

    class Meta:
        model = core_models.User
        fields = (
            "pk", "username", "first_name", "last_name", "is_active",
            "master_user", "mailbox", "role", "language", "phone_number",
            "secondary_email", "domains", "tfa_enabled", "date_joined",
            "last_login"
        )

    def __init__(self, *args, **kwargs):
        """Adapt fields to current user."""
        super().__init__(*args, **kwargs)
        request = self.context.get("request")
        if not request:
            return
        user = self.context["request"].user
        if not user.is_superuser:
            del self.fields["master_user"]

    @extend_schema_field(OpenApiTypes.STR)
    def get_role(self, account):
        """Return role."""
        return account.role

    def get_domains(self, account) -> List[str]:
        """Return domains administered by this account."""
        if account.role not in ["DomainAdmins", "Resellers"]:
            return []
        return admin_models.Domain.objects.get_for_admin(account).values_list(
            "name", flat=True)


class AccountExistsSerializer(serializers.Serializer):
    """Simple serializer used with existence checks."""

    exists = serializers.BooleanField()


class AccountPasswordSerializer(serializers.ModelSerializer):

    """A serializer used to change a user password."""

    new_password = serializers.CharField()

    class Meta:
        model = core_models.User
        fields = (
            "password", "new_password", )

    def validate_password(self, value):
        """Check password."""
        if not self.instance.check_password(value):
            raise serializers.ValidationError("Password not correct")
        return value

    def validate_new_password(self, value):
        """Check new password."""
        try:
            password_validation.validate_password(value, self.instance)
        except ValidationError as exc:
            raise serializers.ValidationError(exc.messages[0])
        return value

    def update(self, instance, validated_data):
        """Set new password."""
        instance.set_password(validated_data["new_password"])
        instance.save()
        return instance


class WritableAccountSerializer(AccountSerializer):
    """Serializer to create account."""

    random_password = serializers.BooleanField(default=False)
    role = serializers.ChoiceField(choices=core_constants.ROLES)
    password = serializers.CharField(required=False)

    class Meta(AccountSerializer.Meta):
        fields = AccountSerializer.Meta.fields + (
            "password", "random_password")

    def __init__(self, *args, **kwargs):
        """Adapt fields to current user."""
        super().__init__(*args, **kwargs)
        request = self.context.get("request")
        if not request:
            return
        user = self.context["request"].user
        self.fields["role"] = serializers.ChoiceField(
            choices=permissions.get_account_roles(user))
        self.fields["domains"] = serializers.ListField(
            child=serializers.CharField(), allow_empty=False, required=False)

    def validate_username(self, value):
        """Lower case username."""
        return value.lower()

    def validate(self, data):
        """Check constraints."""
        master_user = data.get("master_user", False)
        role = data.get("role")
        if master_user and role != "SuperAdmins":
            raise serializers.ValidationError({
                "master_user": _("Not allowed for this role.")
            })
        if role == "SimpleUsers":
            username = data.get("username")
            if username:
                try:
                    validators.UTF8EmailValidator()(username)
                except ValidationError as err:
                    raise ValidationError({"username": err.message})
            mailbox = data.get("mailbox")
            if mailbox is None:
                if not self.instance:
                    data["mailbox"] = {
                        "full_address": username,
                        "use_domain_quota": True
                    }
            elif "full_address" in mailbox and mailbox["full_address"] != username:
                raise serializers.ValidationError({
                    "username": _("Must be equal to mailbox full_address")
                })
        if "mailbox" in data:
            self.address, domain_name = email_utils.split_mailbox(data["mailbox"]["full_address"])
            self.domain = get_object_or_404(
                admin_models.Domain, name=domain_name)
            creator = self.context["request"].user
            if not creator.can_access(self.domain):
                raise serializers.ValidationError({"mailbox": _("Permission denied.")})
            if not self.instance:
                try:
                    core_signals.can_create_object.send(
                        sender=self.__class__, context=creator,
                        klass=admin_models.Mailbox)
                    core_signals.can_create_object.send(
                        sender=self.__class__, context=self.domain,
                        object_type="mailboxes")
                except lib_exceptions.ModoboaException as inst:
                    raise serializers.ValidationError({
                        "mailbox": force_text(inst)})
        condition = (
            not data.get("random_password") and (
                data.get("password") or
                not self.partial
            )
        )
        if condition:
            password = data.get("password")
            if password:
                try:
                    password_validation.validate_password(
                        data["password"], self.instance)
                except ValidationError as exc:
                    raise serializers.ValidationError({
                        "password": exc.messages[0]})
            elif not self.instance:
                raise serializers.ValidationError({
                    "password": _("This field is required.")
                })
        domain_names = data.get("domains")
        if not domain_names:
            return data
        domains = []
        for name in domain_names:
            domain = admin_models.Domain.objects.filter(name=name).first()
            if domain:
                domains.append(domain)
                continue
            raise serializers.ValidationError({
                "domains": _("Local domain {} does not exist").format(name)
            })
        data["domains"] = domains
        return data

    def _create_mailbox(self, creator, account, data):
        """Create a new Mailbox instance."""
        full_address = data.pop("full_address")
        quota = data.pop("quota", None)
        mb = admin_models.Mailbox(
            user=account, address=self.address, domain=self.domain, **data)
        mb.set_quota(quota, creator.has_perm("admin.add_domain"))
        default_msg_limit = param_tools.get_global_parameter(
            "default_mailbox_message_limit")
        if default_msg_limit is not None:
            mb.message_limit = default_msg_limit
        mb.save(creator=creator)
        account.email = full_address
        return mb

    def set_permissions(self, account, domains):
        """Assign permissions on domain(s)."""
        if account.role not in ["DomainAdmins", "Resellers"]:
            return
        current_domains = admin_models.Domain.objects.get_for_admin(account)
        for domain in current_domains:
            if domain not in domains:
                domain.remove_admin(account)
            else:
                domains.remove(domain)
        for domain in domains:
            domain.add_admin(account)

    def create(self, validated_data):
        """Create appropriate objects."""
        creator = self.context["request"].user
        mailbox_data = validated_data.pop("mailbox", None)
        role = validated_data.pop("role")
        domains = validated_data.pop("domains", [])
        random_password = validated_data.pop("random_password", False)
        user = core_models.User(**validated_data)
        if random_password:
            password = lib.make_password()
        else:
            password = validated_data.pop("password")
        user.set_password(password)
        if "language" not in validated_data:
            user.language = settings.LANGUAGE_CODE
        user.save(creator=creator)
        if mailbox_data:
            self._create_mailbox(creator, user, mailbox_data)
        user.role = role
        self.set_permissions(user, domains)
        return user

    def update(self, instance, validated_data):
        """Update account and associated objects."""
        mailbox_data = validated_data.pop("mailbox", None)
        password = validated_data.pop("password", None)
        domains = validated_data.pop("domains", [])
        for key, value in validated_data.items():
            setattr(instance, key, value)
        if password:
            instance.set_password(password)
        if mailbox_data:
            creator = self.context["request"].user
            if instance.mailbox:
                if "full_address" in mailbox_data:
                    # FIXME: compat, to remove ASAP.
                    mailbox_data["email"] = mailbox_data["full_address"]
                    instance.email = mailbox_data["full_address"]
                instance.mailbox.update_from_dict(creator, mailbox_data)
            else:
                self._create_mailbox(creator, instance, mailbox_data)
        instance.save()
        self.set_permissions(instance, domains)
        return instance


class AliasSerializer(serializers.ModelSerializer):
    """Alias serializer."""

    address = lib_fields.DRFEmailFieldUTF8AndEmptyUser()
    recipients = serializers.ListField(
        child=lib_fields.DRFEmailFieldUTF8AndEmptyUser(),
        allow_empty=False,
        help_text=ugettext_lazy("A list of recipient")
    )

    class Meta:
        model = admin_models.Alias
        fields = (
            "pk", "address", "enabled", "internal", "recipients"
        )

    def validate_address(self, value):
        """Check domain."""
        local_part, self.domain = admin_models.validate_alias_address(
            value, self.context["request"].user, instance=self.instance)
        return value

    def create(self, validated_data):
        """Create appropriate objects."""
        creator = self.context["request"].user
        return admin_models.Alias.objects.create(
            creator=creator, domain=self.domain, **validated_data)

    def update(self, instance, validated_data):
        """Update objects."""
        recipients = validated_data.pop("recipients", None)
        for key, value in validated_data.items():
            setattr(instance, key, value)
        instance.save()
        instance.set_recipients(recipients)
        return instance


class SenderAddressSerializer(serializers.ModelSerializer):
    """Base Alias serializer."""

    address = lib_fields.DRFEmailFieldUTF8AndEmptyUser()

    class Meta:
        model = admin_models.SenderAddress
        fields = ("pk", "address", "mailbox")

    def validate_address(self, value):
        """Check domain."""
        local_part, domain = email_utils.split_mailbox(value)
        domain = admin_models.Domain.objects.filter(name=domain).first()
        user = self.context["request"].user
        if domain and not user.can_access(domain):
            raise serializers.ValidationError(
                _("You don't have access to this domain."))
        return value

    def validate_mailbox(self, value):
        """Check permission."""
        user = self.context["request"].user
        if not user.can_access(value):
            raise serializers.ValidationError(
                _("You don't have access to this mailbox."))
        return value


class ResetPasswordSerializer(serializers.Serializer):
    """Serializer by the reset password endpoint."""

    email = lib_fields.DRFEmailFieldUTF8()



"""Management command to check and fix known problems."""

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.encoding import smart_str

from modoboa.admin import models
from modoboa.core.models import User
from modoboa.lib.permissions import get_object_owner, grant_access_to_object

known_problems = []


def known_problem(func):
    """Simple decorator to register a known problem."""
    known_problems.append(func)
    return func


def log(message, quiet=False, **options):
    if not quiet:
        print(message)


def fix_owner(qs, dry_run=False, **options):
    """Fix ownership for orphan objects."""
    model = qs.model
    for obj in qs:
        kw = {"cls": model.__name__, "obj": obj}
        if get_object_owner(obj) is not None:
            continue
        if dry_run:
            log("  {cls} {obj} has no owner".format(**kw), **options)
            continue
        if isinstance(obj, User):
            admin = User.objects.filter(
                is_superuser=True, is_active=True).first()
        elif isinstance(obj, models.Domain):
            admin = obj.admins.first()
        elif isinstance(obj, models.DomainAlias):
            admin = obj.target.admins.first()
        else:
            admin = obj.domain.admins.first()
        if not admin:
            # Fallback: use the first superuser found
            admin = User.objects.filter(
                is_superuser=True, is_active=True).first()
        grant_access_to_object(admin, obj, is_owner=True)
        kw["admin"] = admin
        log("  {cls} {obj} is now owned by {admin}".format(**kw),
            **options)


@known_problem
def sometimes_objects_have_no_owner(**options):
    """Sometime objects have no owner."""
    owned_models = (
        User.objects.all(),
        models.Domain.objects.all(),
        models.DomainAlias.objects.all(),
        models.Alias.objects.filter(domain__isnull=False),
        models.Mailbox.objects.filter(domain__isnull=False),
    )
    for qs in owned_models:
        fix_owner(qs, **options)


@known_problem
def sometimes_mailbox_have_no_alias(**options):
    """Sometime mailboxes have no alias."""
    alias_created = 0
    recipient_created = 0
    for instance in models.Mailbox.objects.select_related("domain").all():
        alias, created = models.Alias.objects.get_or_create(
            address=instance.full_address,
            domain=instance.domain,
            internal=True)
        if created:
            alias_created += 1
            log("Alias {0} created".format(alias), **options)
        recipient, created = models.AliasRecipient.objects.get_or_create(
            alias=alias,
            address=instance.full_address,
            r_mailbox=instance)
        if created:
            recipient_created += 1
            log("AliasRecipient {0} created".format(recipient), **options)
    if alias_created or recipient_created:
        log("{0} alias created. {1} alias recipient created".format(
            alias_created, recipient_created), **options)


class Repair(BaseCommand):
    """Command class."""

    help = "Check and fix known problems."  # NOQA:A003

    def add_arguments(self, parser):
        """Add extra arguments to command."""
        parser.add_argument(
            "--dry-run", action="store_true", default=False,
            help="List known problems without fixing them.")
        parser.add_argument(
            "--quiet", action="store_true", default=False,
            help="Quiet mode.")

    def handle(self, *args, **options):
        """Command entry point."""
        # Load known problems from extensions.
        for ext in settings.MODOBOA_APPS:
            try:
                __import__(
                    ext, locals(), globals(),
                    [smart_str("known_problems")]
                )
            except ImportError:
                pass
        for func in known_problems:
            title = func.__doc__.strip()
            log("", **options)
            log("Checking for... {}...".format(title), **options)
            func(**options)



"""Transport forms."""

from django import forms

from . import backends, models

TYPE_TO_FIELD_MAP = {
    "int": forms.IntegerField,
    "boolean": forms.BooleanField,
    "string": forms.CharField
}


class BackendSettingsMixin(object):
    """A mixin to deal with backend settings in a model form."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super(BackendSettingsMixin, self).__init__(*args, **kwargs)
        self.setting_field_names = []

    def inject_backend_settings(self, name, settings):
        """Inject backend settings to form."""
        for setting in settings:
            fullname = "{}_{}".format(name, setting["name"])
            options = {}
            ftype = setting.get("type", "string")
            if self.instance.pk:
                options["initial"] = self.instance._settings.get(fullname)
            elif "default" in setting:
                options["initial"] = setting["default"]
            if "widget" in setting:
                options["widget"] = setting["widget"]
            self.fields[fullname] = TYPE_TO_FIELD_MAP[ftype](
                label=setting["label"], required=False, **options)
            self.setting_field_names.append(fullname)

    def clean_backend_fields(self, name):
        """Clean backend fields."""
        self.backend = backends.manager.get_backend(name)
        for field, error in self.backend.clean_fields(self.cleaned_data):
            self.add_error(field, error)

    def save(self, commit=True):
        """Set settings to JSON field."""
        transport = super(BackendSettingsMixin, self).save(commit=False)
        transport._settings = {
            name: self.cleaned_data[name]
            for name in self.setting_field_names
        }
        if commit:
            transport.save()
        return transport


class TransportForm(BackendSettingsMixin, forms.ModelForm):
    """Transport model form."""

    service = forms.ChoiceField(choices=[])

    class Meta:
        fields = ("pattern", "service")
        model = models.Transport

    def __init__(self, *args, **kwargs):
        """Set backend list."""
        super(TransportForm, self).__init__(*args, **kwargs)
        self.fields["service"].choices = backends.manager.get_backend_list()
        settings = backends.manager.get_all_backend_settings()
        for name, backend_settings in settings.items():
            self.inject_backend_settings(name, backend_settings)

    @property
    def setting_fields(self):
        return [self[name] for name in self.setting_field_names]

    def _clean_fields(self):
        """Make backend settings required."""
        backend_name = self.data.get("service")
        backend_settings = backends.manager.get_backend_settings(backend_name)
        for name, field in self.fields.items():
            if name.startswith("{}_".format(backend_name)):
                name = name.replace("{}_".format(backend_name), "")
                for setting in backend_settings:
                    if setting["name"] == name:
                        break
                if setting.get("required", True):
                    field.required = True
        return super(TransportForm, self)._clean_fields()

    def clean(self):
        """Check values."""
        cleaned_data = super(TransportForm, self).clean()
        if self.errors:
            return cleaned_data
        self.clean_backend_fields(cleaned_data["service"])
        return cleaned_data



"""Internal library."""

from django.utils.translation import ugettext as _

from modoboa.admin import models as admin_models
from modoboa.lib.exceptions import BadRequest
from modoboa.transport import backends as tr_backends, models as tr_models


def import_relaydomain(user, row, formopts):
    """Specific code for relay domains import"""
    if len(row) != 7:
        raise BadRequest(_("Invalid line"))
    domain = admin_models.Domain(
        name=row[1].strip(), type="relaydomain", quota=0,
        enabled=(row[5].strip().lower() in ["true", "1", "yes", "y"])
    )
    domain.save(creator=user)
    settings = {
        "relay_target_host": row[2].strip(),
        "relay_target_port": row[3].strip(),
        "relay_verify_recipients": (
            row[6].strip().lower() in ["true", "1", "yes", "y"])
    }
    transport = tr_models.Transport(
        pattern=domain.name, service="relay", _settings=settings)
    tr_backends.manager.get_backend("relay").serialize(transport)
    transport.save()



"""Admin API urls."""

from rest_framework import routers

from . import viewsets

router = routers.SimpleRouter()
router.register(r"domains", viewsets.DomainViewSet, basename="domain")
router.register(
    r"domainaliases", viewsets.DomainAliasViewSet, basename="domain_alias")
router.register(r"accounts", viewsets.AccountViewSet, basename="account")
router.register(r"aliases", viewsets.AliasViewSet, basename="alias")
router.register(
    r"senderaddresses", viewsets.SenderAddressViewSet, basename="sender_address")

urlpatterns = router.urls
